from __future__ import annotations

import csv
from pathlib import Path

from babappai.validation.full_pipeline_validation import (
    bootstrap_eii_thresholds,
    confusion_metrics,
    label_positive,
    read_tsv,
)


def _write_metrics_tsv(path: Path) -> None:
    rows = []
    # Balanced tiny dataset: neutral, low, medium, high.
    for i in range(8):
        rows.append({"regime": "neutral", "EII_01": 0.10 + 0.02 * i, "EII_z": -1.5 + 0.1 * i})
    for i in range(4):
        rows.append({"regime": "low", "EII_01": 0.35 + 0.05 * i, "EII_z": -0.2 + 0.2 * i})
    for i in range(4):
        rows.append({"regime": "medium", "EII_01": 0.70 + 0.05 * i, "EII_z": 0.8 + 0.2 * i})
    for i in range(4):
        rows.append({"regime": "high", "EII_01": 0.85 + 0.03 * i, "EII_z": 1.5 + 0.2 * i})

    with path.open("w", newline="") as fh:
        writer = csv.DictWriter(fh, fieldnames=["regime", "EII_01", "EII_z"], delimiter="\t")
        writer.writeheader()
        for row in rows:
            writer.writerow(row)


def test_label_positive_medium_high() -> None:
    labels = label_positive(["neutral", "low", "medium", "high"], "medium_high")
    assert labels.tolist() == [0, 0, 1, 1]


def test_confusion_metrics_sanity() -> None:
    y_true = [0, 0, 1, 1]
    score = [0.1, 0.8, 0.9, 0.2]
    out = confusion_metrics(
        y_true=__import__("numpy").asarray(y_true),
        score=__import__("numpy").asarray(score),
        threshold=0.7,
    )
    assert out["tp"] == 1
    assert out["fp"] == 1
    assert out["tn"] == 1
    assert out["fn"] == 1


def test_bootstrap_eii_thresholds_outputs(tmp_path: Path) -> None:
    metrics_tsv = tmp_path / "metrics.tsv"
    _write_metrics_tsv(metrics_tsv)

    outdir = tmp_path / "bootstrap"
    meta = bootstrap_eii_thresholds(
        metrics_tsv=metrics_tsv,
        outdir=outdir,
        bootstrap_reps=20,
        seed=7,
        decision_target="medium_high",
        default_threshold=0.70,
    )

    summary_rows = read_tsv(meta["bootstrap_summary_tsv"])
    metrics = {row["metric"] for row in summary_rows}
    assert "AUC" in metrics
    assert "FPR_at_default" in metrics
    assert "TPR_at_default" in metrics
    assert "balanced_accuracy_at_default" in metrics



from __future__ import annotations

import os
import re
import subprocess
import sys
from importlib import metadata as importlib_metadata
from pathlib import Path
from tempfile import TemporaryDirectory

import pytest

import babappai.cli as cli
from babappai import __version__


def _read_pyproject_version(pyproject_path: Path) -> str:
    text = pyproject_path.read_text(encoding="utf-8")
    match = re.search(r'(?m)^version\s*=\s*"([^"]+)"\s*$', text)
    if not match:
        raise AssertionError("Could not find static [project].version in pyproject.toml")
    return match.group(1)


def test_pyproject_version_matches_runtime_version():
    pyproject = Path(__file__).resolve().parents[1] / "pyproject.toml"
    assert _read_pyproject_version(pyproject) == __version__


def test_installed_metadata_version_matches_runtime_version():
    cmd = [
        sys.executable,
        "-c",
        "from importlib import metadata; print(metadata.version('babappai'))",
    ]
    with TemporaryDirectory() as temp_dir:
        output = subprocess.check_output(cmd, cwd=temp_dir, text=True).strip()
    if output != __version__ and os.environ.get("CI", "").lower() != "true":
        pytest.skip(
            "Installed distribution version does not match this checkout. "
            "Run after `pip install -e .` to enforce locally."
        )
    assert output == __version__


def test_cli_version_matches_runtime_version(monkeypatch, capsys):
    monkeypatch.setattr(cli, "model_status", lambda: {"cached": True})
    code = cli.main(["version"])
    assert code == 0
    out = capsys.readouterr().out
    assert f"software_version={__version__}" in out



"""Statistical helper utilities for empirical significance workflows."""

from __future__ import annotations

from typing import Any, Iterable, List, Mapping, MutableMapping, Sequence

import numpy as np


def empirical_monte_carlo_pvalue(observed: float, neutral_samples: Sequence[float]) -> float:
    """One-sided empirical Monte Carlo p-value for excess dispersion."""
    arr = np.asarray(list(neutral_samples), dtype=float)
    arr = arr[np.isfinite(arr)]
    m = int(arr.size)
    if m <= 0:
        return 1.0
    exceed = int(np.sum(arr >= float(observed)))
    return float((1 + exceed) / (m + 1))


def bh_adjust(p_values: Sequence[float]) -> np.ndarray:
    """Benjamini-Hochberg FDR adjustment with NaN preservation."""
    arr = np.asarray(p_values, dtype=float)
    out = np.full(arr.shape, np.nan, dtype=float)
    finite_mask = np.isfinite(arr)
    if not np.any(finite_mask):
        return out

    p = arr[finite_mask]
    order = np.argsort(p)
    ranked = p[order]
    m = int(ranked.size)

    ranked_q = np.empty(m, dtype=float)
    running_min = 1.0
    for i in range(m - 1, -1, -1):
        rank = i + 1
        candidate = ranked[i] * m / rank
        running_min = min(running_min, candidate)
        ranked_q[i] = min(running_min, 1.0)

    q = np.empty(m, dtype=float)
    q[order] = ranked_q
    out[finite_mask] = q
    return out


def annotate_bh_qvalues(
    rows: List[MutableMapping[str, Any]],
    *,
    p_key: str = "p_emp",
    q_key: str = "q_emp",
) -> List[MutableMapping[str, Any]]:
    """In-place BH annotation for dict-like rows with p-values."""
    pvals = []
    for row in rows:
        try:
            pvals.append(float(row.get(p_key, float("nan"))))
        except (TypeError, ValueError):
            pvals.append(float("nan"))
    qvals = bh_adjust(pvals)
    for row, q in zip(rows, qvals.tolist()):
        row[q_key] = q
    return rows


__all__ = [
    "annotate_bh_qvalues",
    "bh_adjust",
    "empirical_monte_carlo_pvalue",
]




from __future__ import annotations

import json
from pathlib import Path

import numpy as np

from babappai.calibration.ceii import (
    apply_ceii_calibration,
    binary_metrics,
    derive_threshold,
    expected_calibration_error,
    fit_isotonic_binary,
    load_calibration_asset,
    predict_isotonic,
    save_calibration_asset,
)


def _toy_asset() -> dict:
    return {
        "calibration_version": "ceii_unit_v2",
        "gene_calibrator": {"x": [-2.0, 0.0, 2.0], "y": [0.1, 0.5, 0.9]},
        "site_calibrator": {"x": [-2.0, 0.0, 2.0], "y": [0.05, 0.4, 0.8]},
        "thresholds": {
            "gene": {"threshold": 0.6},
            "site": {"threshold": 0.65},
        },
        "classes": {
            "gene": [
                {"label": "not_identifiable", "min": 0.0, "max": 0.4},
                {"label": "weak_or_ambiguous", "min": 0.4, "max": 0.6},
                {"label": "identifiable", "min": 0.6, "max": 0.8},
                {"label": "strongly_identifiable", "min": 0.8, "max": 1.0},
            ],
            "site": [
                {"label": "not_identifiable", "min": 0.0, "max": 0.4},
                {"label": "weak_or_ambiguous", "min": 0.4, "max": 0.65},
                {"label": "identifiable", "min": 0.65, "max": 0.8},
                {"label": "strongly_identifiable", "min": 0.8, "max": 1.0},
            ],
        },
        "prediction_ci": {
            "gene_lower": {"x": [-2.0, 0.0, 2.0], "y": [0.05, 0.45, 0.85]},
            "gene_upper": {"x": [-2.0, 0.0, 2.0], "y": [0.15, 0.55, 0.95]},
            "site_lower": {"x": [-2.0, 0.0, 2.0], "y": [0.02, 0.35, 0.75]},
            "site_upper": {"x": [-2.0, 0.0, 2.0], "y": [0.08, 0.45, 0.85]},
        },
        "applicability": {
            "features": {
                "n_taxa": {"min": 4, "max": 64},
                "gene_length_nt": {"min": 150, "max": 3000},
            },
            "near_boundary_fraction": 0.08,
            "min_applicability_score_for_calibration": 0.95,
            "allow_near_boundary_calibration": False,
        },
    }


def _toy_linear_asset() -> dict:
    return {
        "calibration_version": "ceii_unit_v2_linear",
        "gene_model": {
            "type": "linear_score_isotonic",
            "feature_names": ["eii_z_raw", "log1p_n_taxa", "log1p_gene_length_nt"],
            "feature_mean": [0.0, 2.5, 7.0],
            "feature_scale": [1.0, 0.5, 0.8],
            "coef": [0.8, 0.2, 0.1],
            "intercept": 0.0,
            "isotonic_calibrator": {"x": [-2.0, 0.0, 2.0], "y": [0.1, 0.5, 0.9]},
            "logistic_calibrator": {"a": 1.0, "b": 0.0},
            "blend_weight_isotonic": 0.5,
            "prediction_ci": {
                "lower": {"x": [-2.0, 0.0, 2.0], "y": [0.05, 0.40, 0.80]},
                "upper": {"x": [-2.0, 0.0, 2.0], "y": [0.15, 0.60, 0.95]},
            },
        },
        "site_model": {
            "type": "linear_score_isotonic",
            "feature_names": ["eii_z_raw", "log1p_n_taxa", "log1p_gene_length_nt"],
            "feature_mean": [0.0, 2.5, 7.0],
            "feature_scale": [1.0, 0.5, 0.8],
            "coef": [0.5, 0.1, 0.05],
            "intercept": -0.2,
            "isotonic_calibrator": {"x": [-2.0, 0.0, 2.0], "y": [0.05, 0.35, 0.7]},
            "logistic_calibrator": {"a": 0.8, "b": -0.1},
            "blend_weight_isotonic": 0.4,
            "prediction_ci": {
                "lower": {"x": [-2.0, 0.0, 2.0], "y": [0.02, 0.25, 0.60]},
                "upper": {"x": [-2.0, 0.0, 2.0], "y": [0.10, 0.45, 0.80]},
            },
        },
        "gene_calibrator": {"x": [-2.0, 0.0, 2.0], "y": [0.1, 0.5, 0.9]},
        "site_calibrator": {"x": [-2.0, 0.0, 2.0], "y": [0.05, 0.35, 0.7]},
        "thresholds": {
            "gene": {"threshold": 0.6},
            "site": {"threshold": 0.5},
        },
        "classes": {
            "gene": [
                {"label": "not_identifiable", "min": 0.0, "max": 0.4},
                {"label": "weak_or_ambiguous", "min": 0.4, "max": 0.6},
                {"label": "identifiable", "min": 0.6, "max": 0.8},
                {"label": "strongly_identifiable", "min": 0.8, "max": 1.0},
            ],
            "site": [
                {"label": "not_identifiable", "min": 0.0, "max": 0.35},
                {"label": "weak_or_ambiguous", "min": 0.35, "max": 0.5},
                {"label": "identifiable", "min": 0.5, "max": 0.75},
                {"label": "strongly_identifiable", "min": 0.75, "max": 1.0},
            ],
        },
        "applicability": {
            "features": {
                "n_taxa": {"min": 8, "max": 64},
                "gene_length_nt": {"min": 120, "max": 4500},
            },
            "near_boundary_fraction": 0.08,
            "min_applicability_score_for_calibration": 0.95,
            "allow_near_boundary_calibration": False,
        },
    }


def test_isotonic_fit_monotone_predictions() -> None:
    x = np.asarray([-1.0, -0.5, 0.0, 0.5, 1.0], dtype=float)
    y = np.asarray([0, 0, 1, 0, 1], dtype=int)
    cal = fit_isotonic_binary(x, y)
    pred = predict_isotonic(cal, x)
    assert np.all(np.diff(pred) >= -1e-12)
    assert np.all((pred >= 0.0) & (pred <= 1.0))


def test_threshold_metrics_and_ece() -> None:
    y = np.asarray([0, 0, 1, 1, 1, 0], dtype=int)
    p = np.asarray([0.1, 0.2, 0.6, 0.7, 0.8, 0.4], dtype=float)
    thr = derive_threshold(y, p, target_fdr=0.2)
    assert 0.0 <= float(thr["threshold"]) <= 1.0
    bm = binary_metrics(y, p, float(thr["threshold"]))
    assert bm["tp"] + bm["fp"] + bm["tn"] + bm["fn"] == int(y.size)
    ece = expected_calibration_error(y, p, n_bins=5)
    assert 0.0 <= float(ece) <= 1.0


def test_apply_ceii_calibration_fields(tmp_path: Path) -> None:
    asset = _toy_asset()
    out = apply_ceii_calibration(
        eii_z_raw=1.0,
        n_taxa=12,
        gene_length_nt=450,
        asset=asset,
    )
    assert 0.0 <= float(out["ceii_gene"]) <= 1.0
    assert 0.0 <= float(out["ceii_site"]) <= 1.0
    assert out["ceii_gene_class"] in {
        "not_identifiable",
        "weak_or_ambiguous",
        "identifiable",
        "strongly_identifiable",
    }
    assert out["domain_shift_or_applicability"] == "in_domain"
    assert out["applicability_status"] == "in_domain"
    assert bool(out["within_applicability_envelope"]) is True
    assert out["calibration_unavailable_reason"] == ""


def test_apply_ceii_calibration_abstains_out_of_domain(tmp_path: Path) -> None:
    asset = _toy_asset()
    out = apply_ceii_calibration(
        eii_z_raw=1.0,
        n_taxa=128,
        gene_length_nt=7000,
        asset=asset,
    )
    assert out["ceii_gene"] is None
    assert out["ceii_site"] is None
    assert out["ceii_gene_class"] == "calibration_unavailable"
    assert out["ceii_site_class"] == "calibration_unavailable"
    assert out["applicability_status"] == "out_of_domain"
    assert bool(out["within_applicability_envelope"]) is False
    assert isinstance(out["calibration_unavailable_reason"], str) and out["calibration_unavailable_reason"]

    # Ensure asset serialization is stable for packaged calibration data.
    path = save_calibration_asset(asset, tmp_path / "asset.json")
    payload = json.loads(path.read_text())
    assert payload["calibration_version"] == "ceii_unit_v2"


def test_apply_ceii_calibration_linear_score_model() -> None:
    asset = _toy_linear_asset()
    out = apply_ceii_calibration(
        eii_z_raw=0.75,
        n_taxa=20,
        gene_length_nt=1200,
        n_branches=37,
        asset=asset,
    )
    assert out["applicability_status"] == "in_domain"
    assert out["ceii_gene"] is not None
    assert out["ceii_site"] is not None
    assert 0.0 <= float(out["ceii_gene"]) <= 1.0
    assert 0.0 <= float(out["ceii_site"]) <= 1.0


def test_packaged_ceii_v2_asset_abstains_out_of_domain() -> None:
    asset = load_calibration_asset()
    assert str(asset.get("calibration_version", "")).startswith("ceii_v2")
    out = apply_ceii_calibration(
        eii_z_raw=0.5,
        n_taxa=512,
        gene_length_nt=9000,
        asset=asset,
    )
    assert out["ceii_gene"] is None
    assert out["ceii_site"] is None
    assert out["ceii_gene_class"] == "calibration_unavailable"
    assert out["ceii_site_class"] == "calibration_unavailable"
    assert str(out["applicability_status"]) in {"out_of_domain", "near_boundary"}



"""Canonical metadata for BABAPPAi and legacy frozen assets.

The package version source of truth is ``[project].version`` in ``pyproject.toml``.
"""

from __future__ import annotations

import re
from importlib import metadata as importlib_metadata
from pathlib import Path

PACKAGE_NAME = "babappai"
SOFTWARE_NAME = "BABAPPAi"
SOFTWARE_DOI = "10.5281/zenodo.18520163"

LEGACY_CODEBASE_NAME = "BABAPPAΩ"
LEGACY_MODEL_NAME = "BABAPPAΩ frozen model"
MODEL_TAG = "legacy_frozen"
MODEL_FILE_NAME = "babappaomega.pt"
MODEL_DOI = "10.5281/zenodo.18195869"
MODEL_URL = "https://zenodo.org/records/18195869/files/babappaomega.pt?download=1"
MODEL_SHA256 = "657a662563af31304abcb208fc903d2770a9184632a9bab2095db4c538fed8eb"
MODEL_COMPATIBILITY_NOTE = (
    "BABAPPAi is the renamed continuation of the BABAPPAΩ codebase. "
    "The cached frozen model is a legacy BABAPPAΩ asset used for "
    "backward-compatible inference until BABAPPAi-specific weights are released."
)


def _read_pyproject_version() -> str:
    pyproject_path = Path(__file__).resolve().parents[1] / "pyproject.toml"
    text = pyproject_path.read_text(encoding="utf-8")
    match = re.search(r'(?m)^version\s*=\s*"([^"]+)"\s*$', text)
    if not match:
        raise RuntimeError("Could not find [project].version in pyproject.toml")
    return match.group(1)


def resolve_software_version() -> str:
    pyproject_path = Path(__file__).resolve().parents[1] / "pyproject.toml"
    if pyproject_path.exists():
        return _read_pyproject_version()

    try:
        return importlib_metadata.version(PACKAGE_NAME)
    except importlib_metadata.PackageNotFoundError:
        raise RuntimeError(
            "Could not resolve software version from pyproject.toml or installed package metadata"
        )


SOFTWARE_VERSION = resolve_software_version()



"""Empirical validation workflow for selected orthogroups."""

from __future__ import annotations

import csv
import json
import random
from pathlib import Path
from typing import Any, Dict, Iterable, List, Optional

from Bio import SeqIO
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord

from babappai.run_pipeline import run_and_write_outputs
from babappai.stats import annotate_bh_qvalues
from babappai.validation.orthogroup_qc import orthogroup_id_from_path


def _load_selected_rows(selected_input: str) -> List[Dict[str, str]]:
    path = Path(selected_input).expanduser().resolve()
    if path.is_dir():
        candidate = path / "selected_100_orthogroups.tsv"
        if not candidate.exists():
            raise FileNotFoundError(
                "Expected selected_100_orthogroups.tsv inside input directory: "
                f"{path}"
            )
        path = candidate

    if not path.exists():
        raise FileNotFoundError(f"Selected orthogroup input not found: {path}")

    with path.open() as fh:
        reader = csv.DictReader(fh, delimiter="\t")
        return [dict(row) for row in reader if row]


def _write_tsv(path: Path, rows: List[Dict[str, Any]], fieldnames: List[str]) -> None:
    with path.open("w", newline="") as fh:
        writer = csv.DictWriter(fh, fieldnames=fieldnames, delimiter="\t")
        writer.writeheader()
        for row in rows:
            writer.writerow(row)


def _annotate_empirical_qvalues(rows: List[Dict[str, Any]], alpha: float) -> None:
    annotate_bh_qvalues(rows, p_key="p_emp", q_key="q_emp")
    for row in rows:
        try:
            qv = float(row.get("q_emp", float("nan")))
        except (TypeError, ValueError):
            qv = float("nan")
        row["alpha_used"] = float(alpha)
        row["significant_bool"] = bool(qv <= alpha) if qv == qv else False
        row["significance_label"] = "significant" if row["significant_bool"] else "not_significant"


def _write_star_tree_for_alignment(alignment_path: Path, out_tree_path: Path) -> Path:
    records = list(SeqIO.parse(alignment_path, "fasta"))
    if not records:
        raise ValueError(f"Alignment is empty: {alignment_path}")
    leaves = ",".join(f"{rec.id}:0.1" for rec in records)
    out_tree_path.parent.mkdir(parents=True, exist_ok=True)
    out_tree_path.write_text(f"({leaves});\n")
    return out_tree_path


def _subsample_alignment(alignment_path: Path, out_path: Path) -> Optional[Path]:
    records = list(SeqIO.parse(alignment_path, "fasta"))
    if len(records) < 4:
        return None
    sampled = records[:-1]
    out_path.parent.mkdir(parents=True, exist_ok=True)
    SeqIO.write(sampled, out_path, "fasta")
    return out_path


def _perturb_alignment(alignment_path: Path, out_path: Path, seed: int) -> Path:
    rng = random.Random(seed)
    records = list(SeqIO.parse(alignment_path, "fasta"))
    if not records:
        raise ValueError(f"Alignment is empty: {alignment_path}")

    codon_choices = [
        "TTT", "TTC", "TTA", "TTG", "CTT", "CTC", "CTA", "CTG",
        "ATT", "ATC", "ATA", "ATG", "GTT", "GTC", "GTA", "GTG",
        "TCT", "TCC", "TCA", "TCG", "CCT", "CCC", "CCA", "CCG",
        "ACT", "ACC", "ACA", "ACG", "GCT", "GCC", "GCA", "GCG",
        "TAT", "TAC", "CAT", "CAC", "CAA", "CAG", "AAT", "AAC",
        "AAA", "AAG", "GAT", "GAC", "GAA", "GAG", "TGT", "TGC",
        "TGG", "CGT", "CGC", "CGA", "CGG", "AGT", "AGC", "AGA",
        "AGG", "GGT", "GGC", "GGA", "GGG",
    ]

    perturbed: List[SeqRecord] = []
    for rec in records:
        seq = str(rec.seq).upper()
        if len(seq) < 3:
            perturbed.append(rec)
            continue
        n_codons = len(seq) // 3
        n_mut = max(1, n_codons // 100)
        positions = sorted(rng.sample(range(n_codons), k=min(n_mut, n_codons)))
        chars = list(seq)
        for pos in positions:
            start = pos * 3
            replacement = rng.choice(codon_choices)
            chars[start:start + 3] = list(replacement)
        perturbed.append(SeqRecord(Seq("".join(chars)), id=rec.id, description=""))

    out_path.parent.mkdir(parents=True, exist_ok=True)
    SeqIO.write(perturbed, out_path, "fasta")
    return out_path


def run_empirical_validation(
    *,
    selected_input: str,
    outdir: str,
    tree_calibration: bool,
    n_calibration: int,
    device: str,
    batch_size: int,
    seed: int,
    foreground_mode: str,
    foreground_list: Optional[str],
    offline: bool,
    overwrite: bool,
    sigma_floor: float,
    alpha: float,
    pvalue_mode: str,
    min_neutral_group_size: int,
    neutral_reps: int,
    robustness_limit: int = 10,
) -> Dict[str, Any]:
    out = Path(outdir).expanduser().resolve()
    out.mkdir(parents=True, exist_ok=True)

    selected_rows = _load_selected_rows(selected_input)
    summary_rows: List[Dict[str, Any]] = []

    for idx, row in enumerate(selected_rows):
        alignment_path = Path(row["alignment_path"]).expanduser().resolve()
        if not alignment_path.exists():
            raise FileNotFoundError(f"Alignment not found: {alignment_path}")

        orthogroup_id = row.get("orthogroup_id") or orthogroup_id_from_path(alignment_path)
        orthogroup_out = out / "orthogroups" / orthogroup_id

        tree_value = row.get("tree_path", "").strip()
        if tree_value and Path(tree_value).exists():
            tree_path = Path(tree_value).resolve()
        else:
            tree_path = _write_star_tree_for_alignment(
                alignment_path,
                out / "generated_trees" / f"{orthogroup_id}.nwk",
            )

        payload = run_and_write_outputs(
            alignment_path=str(alignment_path),
            tree_path=str(tree_path),
            outdir=orthogroup_out,
            command=(
                "babappai validate orthogroups run "
                f"--input {selected_input} --outdir {outdir}"
            ),
            tree_calibration=tree_calibration,
            n_calibration=n_calibration,
            device=device,
            batch_size=batch_size,
            seed=seed,
            foreground_mode=foreground_mode,
            foreground_list=foreground_list,
            offline=offline,
            overwrite=overwrite,
            sigma_floor=sigma_floor,
            alpha=alpha,
            pvalue_mode=pvalue_mode,
            min_neutral_group_size=min_neutral_group_size,
            neutral_reps=neutral_reps,
        )

        gene = payload["gene_summary"]
        summary = {
            "orthogroup_id": orthogroup_id,
            "alignment_path": str(alignment_path),
            "tree_path": str(tree_path),
            "D_obs": gene.get("D_obs"),
            "mu0": gene.get("mu0"),
            "sigma0_raw": gene.get("sigma0_raw"),
            "sigma0_final": gene.get("sigma0_final"),
            "sigma_floor_used": gene.get("sigma_floor_used"),
            "fallback_flag": gene.get("fallback_flag"),
            "fallback_reason": gene.get("fallback_reason"),
            "neutral_group_size": gene.get("neutral_group_size"),
            "calibration_group": gene.get("calibration_group"),
            "calibration_version": gene.get("calibration_version"),
            "applicability_score": gene.get("applicability_score"),
            "applicability_status": gene.get("applicability_status"),
            "within_applicability_envelope": gene.get("within_applicability_envelope"),
            "calibration_unavailable_reason": gene.get("calibration_unavailable_reason"),
            "nearest_supported_regime": gene.get("nearest_supported_regime"),
            "distance_to_supported_domain": gene.get("distance_to_supported_domain"),
            "domain_shift_or_applicability": gene.get("domain_shift_or_applicability"),
            "EII_z": gene["EII_z"],
            "EII_01": gene["EII_01"],
            "eii_z_raw": gene.get("eii_z_raw", gene["EII_z"]),
            "eii_01_raw": gene.get("eii_01_raw", gene["EII_01"]),
            "ceii_gene": gene.get("ceii_gene"),
            "ceii_site": gene.get("ceii_site"),
            "ceii_gene_class": gene.get("ceii_gene_class"),
            "ceii_site_class": gene.get("ceii_site_class"),
            "ceii_gene_identifiable_bool": gene.get("ceii_gene_identifiable_bool"),
            "ceii_site_identifiable_bool": gene.get("ceii_site_identifiable_bool"),
            "p_emp": gene.get("p_emp"),
            "q_emp": gene.get("q_emp"),
            "alpha_used": gene.get("alpha_used"),
            "significant_bool": gene.get("significant_bool"),
            "significance_label": gene.get("significance_label"),
            "identifiable_bool": gene["identifiable_bool"],
            "identifiability_extent": gene["identifiability_extent"],
            "results_json": str(orthogroup_out / "results.json"),
            "branch_summary_tsv": str(orthogroup_out / "branch_summary.tsv"),
            "site_summary_tsv": str(orthogroup_out / "site_summary.tsv"),
            "run_metadata_json": str(orthogroup_out / "run_metadata.json"),
            "repeatability_delta_EII_01": None,
            "calibration_delta_EII_01": None,
            "taxon_subsample_delta_EII_01": None,
            "perturbation_delta_EII_01": None,
        }

        if idx < robustness_limit:
            repeat = run_and_write_outputs(
                alignment_path=str(alignment_path),
                tree_path=str(tree_path),
                outdir=orthogroup_out / "robustness_repeatability",
                command=f"babappai robustness repeatability {orthogroup_id}",
                tree_calibration=tree_calibration,
                n_calibration=n_calibration,
                device=device,
                batch_size=batch_size,
                seed=seed,
                foreground_mode=foreground_mode,
                foreground_list=foreground_list,
                offline=offline,
                overwrite=True,
                sigma_floor=sigma_floor,
                alpha=alpha,
                pvalue_mode=pvalue_mode,
                min_neutral_group_size=min_neutral_group_size,
                neutral_reps=neutral_reps,
            )
            summary["repeatability_delta_EII_01"] = abs(
                float(repeat["gene_summary"]["EII_01"]) - float(gene["EII_01"])
            )

            compare_cal = run_and_write_outputs(
                alignment_path=str(alignment_path),
                tree_path=str(tree_path),
                outdir=orthogroup_out / "robustness_calibration_compare",
                command=f"babappai robustness calibration {orthogroup_id}",
                tree_calibration=not tree_calibration,
                n_calibration=max(20, min(100, n_calibration)),
                device=device,
                batch_size=batch_size,
                seed=seed,
                foreground_mode=foreground_mode,
                foreground_list=foreground_list,
                offline=offline,
                overwrite=True,
                sigma_floor=sigma_floor,
                alpha=alpha,
                pvalue_mode=pvalue_mode,
                min_neutral_group_size=min_neutral_group_size,
                neutral_reps=neutral_reps,
            )
            summary["calibration_delta_EII_01"] = abs(
                float(compare_cal["gene_summary"]["EII_01"]) - float(gene["EII_01"])
            )

            subsampled_alignment = _subsample_alignment(
                alignment_path,
                orthogroup_out / "robustness_taxon_subsample" / "alignment.fasta",
            )
            if subsampled_alignment is not None:
                subsampled_tree = _write_star_tree_for_alignment(
                    subsampled_alignment,
                    orthogroup_out / "robustness_taxon_subsample" / "tree.nwk",
                )
                subsampled = run_and_write_outputs(
                    alignment_path=str(subsampled_alignment),
                    tree_path=str(subsampled_tree),
                    outdir=orthogroup_out / "robustness_taxon_subsample" / "run",
                    command=f"babappai robustness taxon_subsample {orthogroup_id}",
                    tree_calibration=tree_calibration,
                    n_calibration=n_calibration,
                    device=device,
                    batch_size=batch_size,
                    seed=seed,
                    foreground_mode=foreground_mode,
                    foreground_list=foreground_list,
                    offline=offline,
                    overwrite=True,
                    sigma_floor=sigma_floor,
                    alpha=alpha,
                    pvalue_mode=pvalue_mode,
                    min_neutral_group_size=min_neutral_group_size,
                    neutral_reps=neutral_reps,
                )
                summary["taxon_subsample_delta_EII_01"] = abs(
                    float(subsampled["gene_summary"]["EII_01"]) - float(gene["EII_01"])
                )

            perturbed_alignment = _perturb_alignment(
                alignment_path,
                orthogroup_out / "robustness_perturbation" / "alignment.fasta",
                seed=seed + idx + 17,
            )
            perturbed_tree = _write_star_tree_for_alignment(
                perturbed_alignment,
                orthogroup_out / "robustness_perturbation" / "tree.nwk",
            )
            perturbed = run_and_write_outputs(
                alignment_path=str(perturbed_alignment),
                tree_path=str(perturbed_tree),
                outdir=orthogroup_out / "robustness_perturbation" / "run",
                command=f"babappai robustness perturbation {orthogroup_id}",
                tree_calibration=tree_calibration,
                n_calibration=n_calibration,
                device=device,
                batch_size=batch_size,
                seed=seed,
                foreground_mode=foreground_mode,
                foreground_list=foreground_list,
                offline=offline,
                overwrite=True,
                sigma_floor=sigma_floor,
                alpha=alpha,
                pvalue_mode=pvalue_mode,
                min_neutral_group_size=min_neutral_group_size,
                neutral_reps=neutral_reps,
            )
            summary["perturbation_delta_EII_01"] = abs(
                float(perturbed["gene_summary"]["EII_01"]) - float(gene["EII_01"])
            )

        summary_rows.append(summary)

    summary_rows.sort(key=lambda row: row["orthogroup_id"])
    _annotate_empirical_qvalues(summary_rows, alpha=alpha)

    fields = [
        "orthogroup_id",
        "alignment_path",
        "tree_path",
        "D_obs",
        "mu0",
        "sigma0_raw",
        "sigma0_final",
        "sigma_floor_used",
        "fallback_flag",
        "fallback_reason",
        "neutral_group_size",
        "calibration_group",
        "calibration_version",
        "applicability_score",
        "applicability_status",
        "within_applicability_envelope",
        "calibration_unavailable_reason",
        "nearest_supported_regime",
        "distance_to_supported_domain",
        "domain_shift_or_applicability",
        "EII_z",
        "EII_01",
        "eii_z_raw",
        "eii_01_raw",
        "ceii_gene",
        "ceii_site",
        "ceii_gene_class",
        "ceii_site_class",
        "ceii_gene_identifiable_bool",
        "ceii_site_identifiable_bool",
        "p_emp",
        "q_emp",
        "alpha_used",
        "significant_bool",
        "significance_label",
        "identifiable_bool",
        "identifiability_extent",
        "repeatability_delta_EII_01",
        "calibration_delta_EII_01",
        "taxon_subsample_delta_EII_01",
        "perturbation_delta_EII_01",
        "results_json",
        "branch_summary_tsv",
        "site_summary_tsv",
        "run_metadata_json",
    ]

    _write_tsv(out / "empirical_summary.tsv", summary_rows, fields)
    (out / "empirical_summary.json").write_text(json.dumps(summary_rows, indent=2) + "\n")

    metadata = {
        "selected_input": str(Path(selected_input).resolve()),
        "n_orthogroups": len(summary_rows),
        "robustness_limit": robustness_limit,
        "alpha_used": float(alpha),
        "pvalue_mode": pvalue_mode,
        "note": "BABAPPAi is the renamed continuation of the BABAPPAΩ codebase.",
    }
    (out / "empirical_metadata.json").write_text(json.dumps(metadata, indent=2) + "\n")
    return metadata



"""Dispersion statistics for raw EII.

The release-facing primary statistic is ``site_logit_winsorized_variance``:
sample variance (ddof=1) of site-level mean branch logits after two-sided
5% winsorization.

Alternative statistics are exposed for controlled validation/ablation only.
"""

from __future__ import annotations

from typing import Iterable, Sequence

import numpy as np

DISPERSION_METHOD_SITE_LOGIT_VARIANCE = "site_logit_variance"
DISPERSION_METHOD_SITE_LOGIT_WINSORIZED_VARIANCE = "site_logit_winsorized_variance"
DISPERSION_METHOD_SITE_LOGIT_CLIPPED_VARIANCE = "site_logit_clipped_variance"
DISPERSION_METHOD_SITE_LOGIT_MAD_SCALED = "site_logit_mad_scaled"

PRIMARY_DISPERSION_METHOD = DISPERSION_METHOD_SITE_LOGIT_WINSORIZED_VARIANCE

SUPPORTED_DISPERSION_METHODS = (
    DISPERSION_METHOD_SITE_LOGIT_VARIANCE,
    DISPERSION_METHOD_SITE_LOGIT_WINSORIZED_VARIANCE,
    DISPERSION_METHOD_SITE_LOGIT_CLIPPED_VARIANCE,
    DISPERSION_METHOD_SITE_LOGIT_MAD_SCALED,
)


def _as_finite_array(values: Sequence[float] | Iterable[float]) -> np.ndarray:
    arr = np.asarray(list(values), dtype=float)
    if arr.size == 0:
        return arr
    return arr[np.isfinite(arr)]


def compute_dispersion(
    site_logit_mean: Sequence[float] | Iterable[float],
    *,
    method: str = PRIMARY_DISPERSION_METHOD,
    winsor_quantile: float = 0.05,
    clip_abs_logit: float = 6.0,
) -> float:
    arr = _as_finite_array(site_logit_mean)
    if arr.size <= 1:
        return 0.0

    mode = str(method)
    if mode not in SUPPORTED_DISPERSION_METHODS:
        raise ValueError(
            "Unsupported dispersion method "
            f"{mode!r}. Supported: {list(SUPPORTED_DISPERSION_METHODS)}"
        )

    if mode == DISPERSION_METHOD_SITE_LOGIT_VARIANCE:
        return float(np.var(arr, ddof=1))

    if mode == DISPERSION_METHOD_SITE_LOGIT_WINSORIZED_VARIANCE:
        q = float(np.clip(winsor_quantile, 0.0, 0.49))
        lo = float(np.quantile(arr, q))
        hi = float(np.quantile(arr, 1.0 - q))
        clipped = np.clip(arr, lo, hi)
        return float(np.var(clipped, ddof=1))

    if mode == DISPERSION_METHOD_SITE_LOGIT_CLIPPED_VARIANCE:
        bound = float(max(clip_abs_logit, 0.1))
        clipped = np.clip(arr, -bound, bound)
        return float(np.var(clipped, ddof=1))

    # DISPERSION_METHOD_SITE_LOGIT_MAD_SCALED
    median = float(np.median(arr))
    mad = float(np.median(np.abs(arr - median)))
    # Convert MAD to Gaussian-equivalent sigma, then square to keep variance units.
    sigma_equiv = 1.4826 * mad
    return float(sigma_equiv * sigma_equiv)


__all__ = [
    "DISPERSION_METHOD_SITE_LOGIT_CLIPPED_VARIANCE",
    "DISPERSION_METHOD_SITE_LOGIT_MAD_SCALED",
    "DISPERSION_METHOD_SITE_LOGIT_VARIANCE",
    "DISPERSION_METHOD_SITE_LOGIT_WINSORIZED_VARIANCE",
    "PRIMARY_DISPERSION_METHOD",
    "SUPPORTED_DISPERSION_METHODS",
    "compute_dispersion",
]



"""Shared run execution and output writing helpers."""

from __future__ import annotations

import csv
import json
import shlex
from datetime import datetime, timezone
from pathlib import Path
from typing import Any, Dict, Iterable, Optional

from babappai import __version__
from babappai.dispersion import PRIMARY_DISPERSION_METHOD
from babappai.inference import run_inference
from babappai.interpret import interpret_results
from babappai.metadata import (
    MODEL_COMPATIBILITY_NOTE,
    MODEL_DOI,
    MODEL_FILE_NAME,
    MODEL_SHA256,
    SOFTWARE_NAME,
)


def utc_timestamp() -> str:
    return datetime.now(timezone.utc).isoformat()


def command_string(argv: Iterable[str]) -> str:
    return " ".join(shlex.quote(part) for part in argv)


def _write_json(path: Path, payload: Dict[str, Any]) -> None:
    path.write_text(json.dumps(payload, indent=2) + "\n")


def _write_tsv(path: Path, fieldnames: list[str], rows: list[dict[str, Any]]) -> None:
    with path.open("w", newline="") as fh:
        writer = csv.DictWriter(fh, fieldnames=fieldnames, delimiter="\t")
        writer.writeheader()
        for row in rows:
            writer.writerow(row)


def run_and_write_outputs(
    *,
    alignment_path: str,
    tree_path: str,
    outdir: Path,
    command: str,
    tree_calibration: bool,
    n_calibration: int,
    device: str,
    batch_size: int,
    seed: Optional[int],
    foreground_mode: str,
    foreground_list: Optional[str],
    offline: bool,
    overwrite: bool,
    sigma_floor: float = 0.0,
    alpha: float = 0.05,
    pvalue_mode: str = "empirical_monte_carlo",
    dispersion_method: str = PRIMARY_DISPERSION_METHOD,
    retain_eii_bands: bool = True,
    report_threshold_bands: bool = True,
    ceii_enabled: bool = True,
    ceii_asset_path: Optional[str] = None,
    min_neutral_group_size: int = 20,
    neutral_reps: int = 200,
    neutral_generator_metadata: Optional[Dict[str, Any]] = None,
) -> Dict[str, Any]:
    outdir.mkdir(parents=True, exist_ok=True)

    outputs = {
        "results_json": outdir / "results.json",
        "branch_summary_tsv": outdir / "branch_summary.tsv",
        "site_summary_tsv": outdir / "site_summary.tsv",
        "interpretation_txt": outdir / "interpretation.txt",
        "run_metadata_json": outdir / "run_metadata.json",
        "neutral_calibration_replicates_tsv": outdir / "neutral_calibration_replicates.tsv",
    }

    if not overwrite:
        existing = [path for path in outputs.values() if path.exists()]
        if existing:
            raise FileExistsError(
                "Output files already exist. Use --overwrite to replace: "
                + ", ".join(str(path) for path in existing)
            )

    result = run_inference(
        alignment_path=alignment_path,
        tree_path=tree_path,
        model_tag="legacy_frozen",
        tree_calibration=tree_calibration,
        N_calibration=n_calibration,
        device=device,
        batch_size=batch_size,
        seed=seed,
        foreground_mode=foreground_mode,
        foreground_list_path=foreground_list,
        offline=offline,
        sigma_floor=sigma_floor,
        alpha=alpha,
        pvalue_mode=pvalue_mode,
        dispersion_method=dispersion_method,
        neutral_reps=neutral_reps,
        min_neutral_group_size=min_neutral_group_size,
        retain_eii_bands=retain_eii_bands,
        report_threshold_bands=report_threshold_bands,
        ceii_enabled=ceii_enabled,
        ceii_asset_path=ceii_asset_path,
    )

    gene = result["gene_level_identifiability"]
    mstatus = result["model_status"]
    neutral_reps_values = [
        float(x)
        for x in (result.get("neutral_calibration_replicates") or gene.get("neutral_replicates") or [])
    ]

    _write_tsv(
        outputs["neutral_calibration_replicates_tsv"],
        ["replicate_index", "D0_neutral"],
        [
            {"replicate_index": idx + 1, "D0_neutral": value}
            for idx, value in enumerate(neutral_reps_values)
        ],
    )

    payload = {
        "software_name": SOFTWARE_NAME,
        "software_version": __version__,
        "command": command,
        "timestamp": utc_timestamp(),
        "model": {
            "file_name": MODEL_FILE_NAME,
            "doi": MODEL_DOI,
            "sha256": MODEL_SHA256,
            "cached_path": mstatus["cached_path"],
            "verified": bool(mstatus["verified"]),
            "compatibility_note": MODEL_COMPATIBILITY_NOTE,
        },
        "input": {
            "alignment_path": str(Path(alignment_path).resolve()),
            "tree_path": str(Path(tree_path).resolve()),
            "n_sequences": int(result["n_sequences"]),
            "alignment_length": int(result["num_sites"]),
        },
        "calibration": {
            "enabled": bool(tree_calibration),
            "n_calibration": int(n_calibration),
            "source": gene.get("calibration_source"),
            "sigma_floor": float(sigma_floor),
            "alpha": float(alpha),
            "pvalue_mode": str(pvalue_mode),
            "dispersion_method": str(dispersion_method),
            "neutral_reps": int(neutral_reps),
            "min_neutral_group_size": int(min_neutral_group_size),
            "retain_eii_bands": bool(retain_eii_bands),
            "report_threshold_bands": bool(report_threshold_bands),
            "ceii_enabled": bool(ceii_enabled),
            "ceii_asset_path": str(ceii_asset_path) if ceii_asset_path else None,
            "neutral_generator": neutral_generator_metadata,
            "neutral_replicates_tsv": str(outputs["neutral_calibration_replicates_tsv"]),
        },
        "gene_summary": {k: v for k, v in gene.items() if k != "neutral_replicates"},
        "branch_results": result["branch_results"],
        "site_results": result["site_results"],
        "warnings": result.get("warnings", []),
        "outputs": {k: str(v) for k, v in outputs.items()},
    }

    _write_json(outputs["results_json"], payload)

    branch_rows = []
    for row in payload["branch_results"]:
        branch_rows.append(
            {
                "branch": row["branch"],
                "background_score": row["background_score"],
                "logit_mean": row["logit_mean"],
                "is_terminal": row["is_terminal"],
                "selected_foreground": row["selected_foreground"],
                "EII_z": payload["gene_summary"]["EII_z"],
                "EII_01": payload["gene_summary"]["EII_01"],
                "eii_z_raw": payload["gene_summary"]["eii_z_raw"],
                "eii_01_raw": payload["gene_summary"]["eii_01_raw"],
                "ceii_gene": payload["gene_summary"]["ceii_gene"],
                "ceii_site": payload["gene_summary"]["ceii_site"],
                "ceii_gene_class": payload["gene_summary"]["ceii_gene_class"],
                "ceii_site_class": payload["gene_summary"]["ceii_site_class"],
                "ceii_gene_identifiable_bool": payload["gene_summary"]["ceii_gene_identifiable_bool"],
                "ceii_site_identifiable_bool": payload["gene_summary"]["ceii_site_identifiable_bool"],
                "calibration_version": payload["gene_summary"]["calibration_version"],
                "applicability_score": payload["gene_summary"].get("applicability_score"),
                "applicability_status": payload["gene_summary"].get("applicability_status"),
                "within_applicability_envelope": payload["gene_summary"].get("within_applicability_envelope"),
                "calibration_unavailable_reason": payload["gene_summary"].get("calibration_unavailable_reason"),
                "nearest_supported_regime": payload["gene_summary"].get("nearest_supported_regime"),
                "distance_to_supported_domain": payload["gene_summary"].get("distance_to_supported_domain"),
                "sigma0_valid": payload["gene_summary"].get("sigma0_valid"),
                "sigma0_floored": payload["gene_summary"].get("sigma0_floored"),
                "fallback_applied": payload["gene_summary"].get("fallback_applied"),
                "domain_shift_or_applicability": payload["gene_summary"]["domain_shift_or_applicability"],
                "identifiable_bool": payload["gene_summary"]["identifiable_bool"],
                "identifiability_extent": payload["gene_summary"]["identifiability_extent"],
                "p_emp": payload["gene_summary"]["p_emp"],
                "q_emp": payload["gene_summary"]["q_emp"],
                "significant_bool": payload["gene_summary"]["significant_bool"],
                "significance_label": payload["gene_summary"]["significance_label"],
            }
        )

    site_rows = []
    for row in payload["site_results"]:
        site_rows.append(
            {
                "site": row["site"],
                "site_score": row["site_score"],
                "site_logit_mean": row["site_logit_mean"],
                "site_logit_var": row["site_logit_var"],
                "EII_z": payload["gene_summary"]["EII_z"],
                "EII_01": payload["gene_summary"]["EII_01"],
                "eii_z_raw": payload["gene_summary"]["eii_z_raw"],
                "eii_01_raw": payload["gene_summary"]["eii_01_raw"],
                "ceii_gene": payload["gene_summary"]["ceii_gene"],
                "ceii_site": payload["gene_summary"]["ceii_site"],
                "ceii_gene_class": payload["gene_summary"]["ceii_gene_class"],
                "ceii_site_class": payload["gene_summary"]["ceii_site_class"],
                "ceii_gene_identifiable_bool": payload["gene_summary"]["ceii_gene_identifiable_bool"],
                "ceii_site_identifiable_bool": payload["gene_summary"]["ceii_site_identifiable_bool"],
                "calibration_version": payload["gene_summary"]["calibration_version"],
                "applicability_score": payload["gene_summary"].get("applicability_score"),
                "applicability_status": payload["gene_summary"].get("applicability_status"),
                "within_applicability_envelope": payload["gene_summary"].get("within_applicability_envelope"),
                "calibration_unavailable_reason": payload["gene_summary"].get("calibration_unavailable_reason"),
                "nearest_supported_regime": payload["gene_summary"].get("nearest_supported_regime"),
                "distance_to_supported_domain": payload["gene_summary"].get("distance_to_supported_domain"),
                "sigma0_valid": payload["gene_summary"].get("sigma0_valid"),
                "sigma0_floored": payload["gene_summary"].get("sigma0_floored"),
                "fallback_applied": payload["gene_summary"].get("fallback_applied"),
                "domain_shift_or_applicability": payload["gene_summary"]["domain_shift_or_applicability"],
                "identifiable_bool": payload["gene_summary"]["identifiable_bool"],
                "identifiability_extent": payload["gene_summary"]["identifiability_extent"],
                "p_emp": payload["gene_summary"]["p_emp"],
                "q_emp": payload["gene_summary"]["q_emp"],
                "significant_bool": payload["gene_summary"]["significant_bool"],
                "significance_label": payload["gene_summary"]["significance_label"],
            }
        )

    _write_tsv(
        outputs["branch_summary_tsv"],
        [
            "branch",
            "background_score",
            "logit_mean",
            "is_terminal",
            "selected_foreground",
            "EII_z",
            "EII_01",
            "eii_z_raw",
            "eii_01_raw",
            "ceii_gene",
            "ceii_site",
            "ceii_gene_class",
            "ceii_site_class",
            "ceii_gene_identifiable_bool",
            "ceii_site_identifiable_bool",
            "calibration_version",
            "applicability_score",
            "applicability_status",
            "within_applicability_envelope",
            "calibration_unavailable_reason",
            "nearest_supported_regime",
            "distance_to_supported_domain",
            "sigma0_valid",
            "sigma0_floored",
            "fallback_applied",
            "domain_shift_or_applicability",
            "identifiable_bool",
            "identifiability_extent",
            "p_emp",
            "q_emp",
            "significant_bool",
            "significance_label",
        ],
        branch_rows,
    )

    _write_tsv(
        outputs["site_summary_tsv"],
        [
            "site",
            "site_score",
            "site_logit_mean",
            "site_logit_var",
            "EII_z",
            "EII_01",
            "eii_z_raw",
            "eii_01_raw",
            "ceii_gene",
            "ceii_site",
            "ceii_gene_class",
            "ceii_site_class",
            "ceii_gene_identifiable_bool",
            "ceii_site_identifiable_bool",
            "calibration_version",
            "applicability_score",
            "applicability_status",
            "within_applicability_envelope",
            "calibration_unavailable_reason",
            "nearest_supported_regime",
            "distance_to_supported_domain",
            "sigma0_valid",
            "sigma0_floored",
            "fallback_applied",
            "domain_shift_or_applicability",
            "identifiable_bool",
            "identifiability_extent",
            "p_emp",
            "q_emp",
            "significant_bool",
            "significance_label",
        ],
        site_rows,
    )

    interpret_results(payload, out_path=outputs["interpretation_txt"])

    run_metadata = {
        "software_name": SOFTWARE_NAME,
        "software_version": __version__,
        "timestamp": utc_timestamp(),
        "command": command,
        "inputs": payload["input"],
        "calibration": payload["calibration"],
        "model": payload["model"],
        "outputs": {k: str(v) for k, v in outputs.items()},
    }
    _write_json(outputs["run_metadata_json"], run_metadata)

    return payload


def terminal_summary(payload: Dict[str, Any]) -> list[str]:
    gene = payload["gene_summary"]
    branches = sorted(
        payload["branch_results"],
        key=lambda row: float(row["background_score"]),
        reverse=True,
    )
    sites = sorted(
        payload["site_results"],
        key=lambda row: float(row["site_score"]),
        reverse=True,
    )

    branch_labels = ", ".join(
        f"{row['branch']} ({float(row['background_score']):.3f})"
        for row in branches[:3]
    ) or "n/a"
    site_labels = ", ".join(
        f"{int(row['site'])}:{float(row['site_score']):.3f}"
        for row in sites[:5]
    ) or "n/a"

    def _fmt_opt_prob(value: Any) -> str:
        if value is None:
            return "n/a"
        try:
            fv = float(value)
        except (TypeError, ValueError):
            return "n/a"
        return f"{fv:.2f}" if fv == fv else "n/a"

    return [
        f"Empirical p-value (p_emp): {float(gene['p_emp']):.4g}" if gene.get("p_emp") is not None else "Empirical p-value (p_emp): n/a",
        f"BH q-value (q_emp): {float(gene['q_emp']):.4g}" if gene.get("q_emp") is not None else "BH q-value (q_emp): n/a",
        f"Significant at alpha={float(gene.get('alpha_used', 0.05)):.2f}: {'YES' if bool(gene.get('significant_bool')) else 'NO'}",
        f"Gene-level raw EII_z: {float(gene['eii_z_raw']):.2f}",
        f"Gene-level raw EII_01: {float(gene['eii_01_raw']):.2f}",
        f"cEII_gene (P[identifiable]): {_fmt_opt_prob(gene.get('ceii_gene'))}",
        f"cEII_site (P[identifiable]): {_fmt_opt_prob(gene.get('ceii_site'))}",
        f"cEII gene class: {gene.get('ceii_gene_class', 'n/a')} (calibration={gene.get('calibration_version', 'unknown')})",
        (
            "Applicability: "
            f"{gene.get('applicability_status', gene.get('domain_shift_or_applicability', 'unknown'))} "
            f"(score={_fmt_opt_prob(gene.get('applicability_score'))})"
        ),
        (
            f"Calibration unavailable reason: {gene.get('calibration_unavailable_reason')}"
            if gene.get("calibration_unavailable_reason")
            else "Calibration unavailable reason: none"
        ),
        (
            "WARNING: calibrated cEII outputs withheld due to applicability/null-calibration checks."
            if gene.get("ceii_gene") is None or gene.get("ceii_site") is None
            else "Calibrated cEII outputs available."
        ),
        (
            f"sigma0_valid={bool(gene.get('sigma0_valid'))}, "
            f"sigma0_floored={bool(gene.get('sigma0_floored'))}, "
            f"fallback_applied={bool(gene.get('fallback_applied', gene.get('fallback_flag')))}"
        ),
        f"Strongest branches: {branch_labels}",
        f"Top elevated sites: {site_labels}",
        "Note: raw EII is a dispersion magnitude; cEII is the calibrated identifiability probability layer.",
        "Significance uses empirical p/q relative to matched neutral calibration.",
        "Warning: significance does not prove adaptive substitution.",
    ]



"""Human-readable interpretation report generation."""

from __future__ import annotations

import json
from pathlib import Path
from typing import Dict, Iterable, List, Mapping


def _load_results(result_input) -> Dict[str, object]:
    if isinstance(result_input, (str, Path)):
        return json.loads(Path(result_input).read_text())
    if isinstance(result_input, Mapping):
        return dict(result_input)
    raise TypeError("result_input must be a mapping or a JSON file path")


def _regime_explanation(regime: str) -> str:
    mapping = {
        "not_identifiable": (
            "Current data do not support reliable episodic branch-site "
            "identifiability at the gene level within the calibrated simulation domain."
        ),
        "weak_or_ambiguous": (
            "Calibrated probability indicates ambiguous recoverability; interpret as "
            "exploratory and seek additional sequence depth or broader taxon sampling."
        ),
        "identifiable": (
            "The gene-level pattern is measurably identifiable under the empirical "
            "calibration regime."
        ),
        "strongly_identifiable": (
            "The gene-level pattern is strongly identifiable under the empirical "
            "calibration regime."
        ),
        "calibration_unavailable": (
            "Calibrated identifiability probability is abstained because the sample "
            "is outside or near the boundary of validated calibration support."
        ),
    }
    return mapping.get(regime, "Interpretation regime unavailable.")


def _sorted_branch_rows(branch_results: Iterable[Mapping[str, object]]) -> List[Mapping[str, object]]:
    return sorted(
        branch_results,
        key=lambda row: float(row.get("background_score", 0.0)),
        reverse=True,
    )


def _sorted_site_rows(site_results: Iterable[Mapping[str, object]]) -> List[Mapping[str, object]]:
    return sorted(
        site_results,
        key=lambda row: float(row.get("site_score", 0.0)),
        reverse=True,
    )


def render_interpretation(result_input, *, top_branches: int = 10, top_sites: int = 20) -> str:
    results = _load_results(result_input)

    gene = results.get("gene_summary") or results.get("gene_level_identifiability", {})
    eii_z_raw = gene.get("eii_z_raw", gene.get("EII_z"))
    eii_01_raw = gene.get("eii_01_raw", gene.get("EII_01"))
    ceii_gene = gene.get("ceii_gene")
    ceii_site = gene.get("ceii_site")
    ceii_gene_class = gene.get("ceii_gene_class", gene.get("identifiability_extent"))
    ceii_site_class = gene.get("ceii_site_class", "unavailable")
    ceii_ci = gene.get("ceii_ci", {})
    calibration_version = gene.get("calibration_version", "unknown")
    applicability = gene.get("domain_shift_or_applicability", "unknown")
    applicability_score = gene.get("applicability_score")
    applicability_status = gene.get("applicability_status", applicability)
    within_envelope = gene.get("within_applicability_envelope")
    calib_unavailable_reason = gene.get("calibration_unavailable_reason")
    nearest_supported_regime = gene.get("nearest_supported_regime")
    distance_to_supported_domain = gene.get("distance_to_supported_domain")
    sigma0_valid = gene.get("sigma0_valid")
    sigma0_floored = gene.get("sigma0_floored")
    fallback_applied = gene.get("fallback_applied", gene.get("fallback_flag"))
    p_emp = gene.get("p_emp")
    q_emp = gene.get("q_emp")
    alpha_used = gene.get("alpha_used")
    significant_bool = gene.get("significant_bool")
    significance_label = gene.get("significance_label")
    identifiable = gene.get("ceii_gene_identifiable_bool", gene.get("identifiable_bool"))
    extent = ceii_gene_class

    branch_results = _sorted_branch_rows(results.get("branch_results", []))
    site_results = _sorted_site_rows(results.get("site_results", []))

    lines: List[str] = []
    lines.append("BABAPPAi Interpretation Report")
    lines.append("=" * 60)
    lines.append("")
    lines.append("Verdict")
    lines.append("-" * 60)
    if p_emp is not None:
        lines.append(f"Empirical p-value (p_emp): {float(p_emp):.4g}")
    if q_emp is not None:
        lines.append(f"BH-adjusted q-value (q_emp): {float(q_emp):.4g}")
    if alpha_used is not None:
        lines.append(
            f"Significant at q <= {float(alpha_used):.2f}: "
            f"{'YES' if bool(significant_bool) else 'NO'} ({significance_label})"
        )
    lines.append(f"Gene-level raw EII_z: {float(eii_z_raw):.3f}")
    lines.append(f"Gene-level raw EII_01: {float(eii_01_raw):.3f}")
    lines.append(f"cEII_gene (P[gene identifiable]): {float(ceii_gene):.3f}" if ceii_gene is not None else "cEII_gene: n/a")
    lines.append(f"cEII_site (P[site identifiable]): {float(ceii_site):.3f}" if ceii_site is not None else "cEII_site: n/a")
    if isinstance(ceii_ci, Mapping):
        gene_ci = ceii_ci.get("gene", {})
        if isinstance(gene_ci, Mapping):
            lo = gene_ci.get("lower")
            hi = gene_ci.get("upper")
            if lo is not None and hi is not None:
                lines.append(f"cEII_gene CI: [{float(lo):.3f}, {float(hi):.3f}]")
    lines.append(f"cEII gene class: {extent}")
    lines.append(f"cEII site class: {ceii_site_class}")
    lines.append(f"cEII gene identifiable at calibrated threshold: {'YES' if bool(identifiable) else 'NO'}")
    lines.append(f"Calibration version: {calibration_version}")
    lines.append(f"Applicability/domain flag: {applicability}")
    lines.append(f"Applicability status: {applicability_status}")
    if applicability_score is not None:
        lines.append(f"Applicability score: {float(applicability_score):.3f}")
    if within_envelope is not None:
        lines.append(f"Within applicability envelope: {'YES' if bool(within_envelope) else 'NO'}")
    if nearest_supported_regime:
        lines.append(f"Nearest supported regime: {nearest_supported_regime}")
    if distance_to_supported_domain is not None:
        lines.append(f"Distance to supported domain: {float(distance_to_supported_domain):.3f}")
    if calib_unavailable_reason:
        lines.append(f"Calibration unavailable reason: {calib_unavailable_reason}")
    if sigma0_valid is not None:
        lines.append(f"sigma0_valid: {'YES' if bool(sigma0_valid) else 'NO'}")
    if sigma0_floored is not None:
        lines.append(f"sigma0_floored: {'YES' if bool(sigma0_floored) else 'NO'}")
    if fallback_applied is not None:
        lines.append(f"fallback_applied: {'YES' if bool(fallback_applied) else 'NO'}")
    lines.append(_regime_explanation(str(extent)))
    lines.append("")

    lines.append("Strongest Branches")
    lines.append("-" * 60)
    for idx, row in enumerate(branch_results[:top_branches], start=1):
        lines.append(
            f"{idx:2d}. {row['branch']:<30} "
            f"score={float(row['background_score']):.4f} "
            f"selected={bool(row.get('selected_foreground', False))}"
        )
    if not branch_results:
        lines.append("No branch-level rows available.")
    lines.append("")

    lines.append("Top Elevated Sites")
    lines.append("-" * 60)
    for row in site_results[:top_sites]:
        lines.append(
            f"Site {int(row['site']):4d} "
            f"score={float(row['site_score']):.4f} "
            f"logit_mean={float(row['site_logit_mean']):+.4f}"
        )
    if not site_results:
        lines.append("No site-level rows available.")
    lines.append("")

    lines.append("Interpretation Policy")
    lines.append("-" * 60)
    lines.append("raw EII_z / EII_01 are dispersion magnitude diagnostics.")
    lines.append("cEII_gene / cEII_site are empirically calibrated probabilities of recoverability.")
    lines.append("cEII class labels are derived from the loaded calibration asset, not fixed constants.")
    lines.append("If applicability fails or null-dispersion calibration is invalid/floored, cEII abstains (null).")
    lines.append("")
    lines.append(
        "Inferential significance is based on empirical p_emp/q_emp from matched neutral calibration; "
        "this is distinct from cEII probability calibration."
    )
    lines.append(
        "Warning: significance indicates excess dispersion relative to matched neutral expectation, "
        "not proof of adaptive substitution."
    )

    return "\n".join(lines) + "\n"


def interpret_results(result_json, out_path=None, top_branches=10, top_sites=20):
    report = render_interpretation(
        result_json,
        top_branches=top_branches,
        top_sites=top_sites,
    )

    if out_path is None:
        if isinstance(result_json, (str, Path)):
            out_path = Path(result_json).with_name("interpretation.txt")
        else:
            raise ValueError("out_path is required when result_json is a mapping")

    out_path = Path(out_path)
    out_path.write_text(report)
    return out_path



"""Legacy raw-EII descriptive helpers.

Primary release-facing identifiability decisions should use calibrated cEII
probabilities. This module remains for backward-compatible descriptive output.
"""

from __future__ import annotations

import math
from typing import Dict, Optional, Tuple


REGIME_NOT_IDENTIFIABLE = "not_identifiable"
REGIME_WEAK_OR_AMBIGUOUS = "weak_or_ambiguous"
REGIME_IDENTIFIABLE = "identifiable"
REGIME_STRONGLY_IDENTIFIABLE = "strongly_identifiable"
EII_BANDS_DESCRIPTIVE_ONLY = True
LEGACY_FALLBACK_THRESHOLDS = (0.30, 0.70, 0.90)


def eii01_from_eiiz(eii_z: float) -> float:
    """Map raw EII_z into bounded [0, 1] space via a stable sigmoid transform."""
    clipped = max(min(eii_z, 60.0), -60.0)
    return 1.0 / (1.0 + math.exp(-clipped))


def _resolve_thresholds(calibration_asset_path: Optional[str] = None) -> Tuple[Tuple[float, float, float], str]:
    try:
        from babappai.calibration import load_calibration_asset

        asset = load_calibration_asset(calibration_asset_path)
        gene_thr = asset.get("thresholds", {}).get("gene", {})
        weak = float(gene_thr.get("weak_threshold", LEGACY_FALLBACK_THRESHOLDS[0]))
        main = float(gene_thr.get("threshold", LEGACY_FALLBACK_THRESHOLDS[1]))
        strong = float(gene_thr.get("strong_threshold", LEGACY_FALLBACK_THRESHOLDS[2]))
        if not (0.0 <= weak <= main <= strong <= 1.0):
            raise ValueError("invalid threshold ordering in calibration asset")
        return (weak, main, strong), "empirical_calibration_asset"
    except Exception:
        return LEGACY_FALLBACK_THRESHOLDS, "legacy_fallback_fixed_thresholds"


def identifiability_extent(eii_01: float, *, thresholds: Tuple[float, float, float] = LEGACY_FALLBACK_THRESHOLDS) -> str:
    weak, main, strong = thresholds
    if 0.0 <= eii_01 < weak:
        return REGIME_NOT_IDENTIFIABLE
    if weak <= eii_01 < main:
        return REGIME_WEAK_OR_AMBIGUOUS
    if main <= eii_01 < strong:
        return REGIME_IDENTIFIABLE
    return REGIME_STRONGLY_IDENTIFIABLE


def identifiability_bool(eii_01: float, *, threshold: float = LEGACY_FALLBACK_THRESHOLDS[1]) -> bool:
    """Legacy descriptive flag retained for backward compatibility."""
    return eii_01 >= float(threshold)


def interpret_identifiability(eii_z: float, *, calibration_asset_path: Optional[str] = None) -> Dict[str, object]:
    eii_01 = eii01_from_eiiz(eii_z)
    thresholds, source = _resolve_thresholds(calibration_asset_path)
    extent = identifiability_extent(eii_01, thresholds=thresholds)
    return {
        "EII_01": eii_01,
        "identifiable_bool": identifiability_bool(eii_01, threshold=thresholds[1]),
        "identifiability_extent": extent,
        "eii_band_descriptive_only": EII_BANDS_DESCRIPTIVE_ONLY,
        "threshold_source": source,
        "threshold_values": {
            "weak": thresholds[0],
            "main": thresholds[1],
            "strong": thresholds[2],
        },
    }



"""Truth-aware recoverability metrics used to calibrate cEII."""

from __future__ import annotations

import csv
import json
import math
from pathlib import Path
from typing import Any, Dict, Iterable, List, Mapping, MutableMapping, Sequence

import numpy as np


def _rankdata(values: np.ndarray) -> np.ndarray:
    order = np.argsort(values, kind="mergesort")
    ranks = np.empty(values.size, dtype=float)
    i = 0
    while i < values.size:
        j = i + 1
        while j < values.size and values[order[j]] == values[order[i]]:
            j += 1
        rank = 0.5 * (i + j - 1) + 1.0
        ranks[order[i:j]] = rank
        i = j
    return ranks


def spearman_corr(x: Sequence[float], y: Sequence[float]) -> float:
    xa = np.asarray(x, dtype=float)
    ya = np.asarray(y, dtype=float)
    mask = np.isfinite(xa) & np.isfinite(ya)
    xa = xa[mask]
    ya = ya[mask]
    if xa.size < 3:
        return float("nan")
    if np.all(xa == xa[0]) or np.all(ya == ya[0]):
        return float("nan")
    rx = _rankdata(xa)
    ry = _rankdata(ya)
    sx = np.std(rx, ddof=1)
    sy = np.std(ry, ddof=1)
    if sx <= 0 or sy <= 0:
        return float("nan")
    return float(np.corrcoef(rx, ry)[0, 1])


def _normalize_rank_metric(rho: float, *, neutral_fallback: float = 0.5) -> float:
    if not np.isfinite(rho):
        return float(neutral_fallback)
    return float(np.clip((rho + 1.0) / 2.0, 0.0, 1.0))


def _read_site_scores(path: str | Path) -> np.ndarray:
    vals: List[float] = []
    with Path(path).open() as fh:
        for row in csv.DictReader(fh, delimiter="\t"):
            vals.append(float(row["site_score"]))
    return np.asarray(vals, dtype=float)


def _read_branch_scores(path: str | Path) -> Dict[str, float]:
    out: Dict[str, float] = {}
    with Path(path).open() as fh:
        for row in csv.DictReader(fh, delimiter="\t"):
            out[str(row["branch"])] = float(row["background_score"])
    return out


def _load_truth_payload(truth_metadata_path: str | Path) -> Dict[str, Any]:
    payload = json.loads(Path(truth_metadata_path).read_text())
    return payload


def _load_latent_truth_arrays(truth_metadata_path: str | Path) -> Dict[str, Any]:
    truth_meta = _load_truth_payload(truth_metadata_path)
    latent_npz = truth_meta.get("latent_truth_npz_path")
    if not latent_npz:
        return {}
    npz_path = Path(latent_npz)
    if not npz_path.exists():
        return {}
    data = np.load(npz_path, allow_pickle=True)
    return {
        "site_burden_true": np.asarray(data["site_burden_true"], dtype=float),
        "branch_burden_true": np.asarray(data["branch_burden_true"], dtype=float),
        "branch_names": [str(x) for x in data["branch_names"].tolist()],
        "active_site_indicator": np.asarray(data["active_site_indicator"], dtype=int),
        "active_branch_indicator": np.asarray(data["active_branch_indicator"], dtype=int),
        "latent_cell_fraction_true": float(np.asarray(data["branch_site_matrix_true"], dtype=float).mean()),
    }


def _site_enrichment_at_k(site_score: np.ndarray, active_site_indicator: np.ndarray) -> float:
    if site_score.size == 0 or active_site_indicator.size == 0:
        return float("nan")
    n = min(site_score.size, active_site_indicator.size)
    score = site_score[:n]
    active = active_site_indicator[:n].astype(int)
    n_pos = int(np.sum(active))
    if n_pos <= 0:
        # Neutral case: reward low top-k signal.
        k = max(1, int(math.ceil(0.10 * n)))
        topk = np.sort(score)[-k:]
        return float(np.clip(1.0 - float(np.mean(topk)), 0.0, 1.0))

    k = max(1, n_pos)
    idx = np.argsort(score)[-k:]
    hit = int(np.sum(active[idx] == 1))
    return float(hit / k)


def _burden_alignment_score(pred_gene_burden: float, true_gene_burden: float) -> float:
    err = abs(float(pred_gene_burden) - float(true_gene_burden))
    scale = 0.05 + 0.5 * max(float(true_gene_burden), 0.0)
    return float(np.exp(-err / scale))


def compute_truth_aware_metrics(row: Mapping[str, Any]) -> Dict[str, float]:
    site_scores = _read_site_scores(row["site_summary_tsv"])
    branch_scores = _read_branch_scores(row["branch_summary_tsv"])
    truth = _load_latent_truth_arrays(row["truth_metadata_path"])

    if not truth:
        # fallback for older datasets lacking latent arrays
        pred_gene_burden = float(np.mean(site_scores)) if site_scores.size > 0 else float("nan")
        true_gene_burden = float(row.get("latent_cell_fraction_realized", 0.0))
        return {
            "site_enrichment_at_k": float("nan"),
            "site_spearman": float("nan"),
            "branch_spearman": float("nan"),
            "pred_gene_burden": pred_gene_burden,
            "true_gene_burden": true_gene_burden,
            "burden_alignment_score": _burden_alignment_score(pred_gene_burden, true_gene_burden),
        }

    site_true = np.asarray(truth["site_burden_true"], dtype=float)
    branch_true = np.asarray(truth["branch_burden_true"], dtype=float)
    active_site = np.asarray(truth["active_site_indicator"], dtype=int)
    branch_names = truth["branch_names"]

    n_site = min(site_scores.size, site_true.size)
    site_scores = site_scores[:n_site]
    site_true = site_true[:n_site]
    active_site = active_site[:n_site]

    pred_branch = np.asarray([float(branch_scores.get(name, 0.0)) for name in branch_names], dtype=float)
    n_branch = min(pred_branch.size, branch_true.size)
    pred_branch = pred_branch[:n_branch]
    branch_true = branch_true[:n_branch]

    site_srho = spearman_corr(site_scores, site_true)
    branch_srho = spearman_corr(pred_branch, branch_true)

    pred_gene_burden = float(np.mean(site_scores)) if site_scores.size > 0 else float("nan")
    true_gene_burden = float(np.mean(site_true)) if site_true.size > 0 else float("nan")

    return {
        "site_enrichment_at_k": _site_enrichment_at_k(site_scores, active_site),
        "site_spearman": site_srho,
        "branch_spearman": branch_srho,
        "pred_gene_burden": pred_gene_burden,
        "true_gene_burden": true_gene_burden,
        "burden_alignment_score": _burden_alignment_score(pred_gene_burden, true_gene_burden),
    }


def attach_scenario_stability(rows: List[MutableMapping[str, Any]]) -> List[MutableMapping[str, Any]]:
    by_scenario: Dict[str, List[MutableMapping[str, Any]]] = {}
    for row in rows:
        by_scenario.setdefault(str(row["scenario_id"]), []).append(row)

    for _sid, scenario_rows in by_scenario.items():
        site_vectors = []
        branch_vectors = []
        for row in scenario_rows:
            site_vectors.append(_read_site_scores(row["site_summary_tsv"]))
            branch_scores = _read_branch_scores(row["branch_summary_tsv"])
            branch_names = sorted(branch_scores.keys())
            branch_vectors.append(np.asarray([branch_scores[b] for b in branch_names], dtype=float))

        def _pairwise_mean_corr(vectors: List[np.ndarray]) -> float:
            vals = []
            for i in range(len(vectors)):
                for j in range(i + 1, len(vectors)):
                    n = min(vectors[i].size, vectors[j].size)
                    if n < 3:
                        continue
                    corr = np.corrcoef(vectors[i][:n], vectors[j][:n])[0, 1]
                    if np.isfinite(corr):
                        vals.append((corr + 1.0) / 2.0)
            if not vals:
                return 1.0
            return float(np.mean(vals))

        site_stability = _pairwise_mean_corr(site_vectors)
        branch_stability = _pairwise_mean_corr(branch_vectors)

        for row in scenario_rows:
            row["scenario_site_stability"] = site_stability
            row["scenario_branch_stability"] = branch_stability
    return rows


def attach_recoverability_targets(
    rows: List[MutableMapping[str, Any]],
    *,
    tau_gene: float = 0.42,
    tau_site: float = 0.45,
) -> List[MutableMapping[str, Any]]:
    for row in rows:
        site_enrich = float(row.get("site_enrichment_at_k", float("nan")))
        site_rank = _normalize_rank_metric(float(row.get("site_spearman", float("nan"))), neutral_fallback=0.5)
        branch_rank = _normalize_rank_metric(float(row.get("branch_spearman", float("nan"))), neutral_fallback=0.5)
        burden_align = float(row.get("burden_alignment_score", float("nan")))
        site_stab = float(row.get("scenario_site_stability", 1.0))
        branch_stab = float(row.get("scenario_branch_stability", 1.0))

        if not np.isfinite(site_enrich):
            site_enrich = 0.5
        if not np.isfinite(burden_align):
            burden_align = 0.5
        if not np.isfinite(site_stab):
            site_stab = 0.5
        if not np.isfinite(branch_stab):
            branch_stab = 0.5

        r_gene = float(np.clip(0.45 * branch_rank + 0.35 * burden_align + 0.20 * branch_stab, 0.0, 1.0))
        r_site = float(np.clip(0.45 * site_enrich + 0.35 * site_rank + 0.20 * site_stab, 0.0, 1.0))

        row["R_gene"] = r_gene
        row["R_site"] = r_site
        row["I_gene"] = int(r_gene >= float(tau_gene))
        row["I_site"] = int(r_site >= float(tau_site))
    return rows


def assign_scenario_splits(
    rows: List[MutableMapping[str, Any]],
    *,
    seed: int = 123,
) -> List[MutableMapping[str, Any]]:
    # OOD: intentionally difficult strata not used for calibration threshold fitting.
    scenario_keys: Dict[str, Dict[str, Any]] = {}
    for row in rows:
        sid = str(row["scenario_id"])
        scenario_keys.setdefault(
            sid,
            {
                "scenario_id": sid,
                "regime": str(row.get("regime", "")),
                "tree_bin": str(row.get("tree_bin", "")),
                "recombination_bin": str(row.get("recombination_bin", "")),
                "alignment_noise_bin": str(row.get("alignment_noise_bin", "")),
                "n_taxa": int(float(row.get("n_taxa", 0) or 0)),
            },
        )

    ood_ids = {
        sid
        for sid, meta in scenario_keys.items()
        if (
            int(meta.get("n_taxa", 0)) >= 24
            and (meta["recombination_bin"] == "high" or meta["alignment_noise_bin"] == "high")
        )
    }

    rng = np.random.default_rng(seed)
    by_regime: Dict[str, List[str]] = {}
    for sid, meta in scenario_keys.items():
        if sid in ood_ids:
            continue
        by_regime.setdefault(meta["regime"], []).append(sid)

    split_map: Dict[str, str] = {}
    for sid in ood_ids:
        split_map[sid] = "ood"

    for _regime, ids in by_regime.items():
        ids = list(ids)
        rng.shuffle(ids)
        n = len(ids)
        n_train = int(round(0.60 * n))
        n_cal = int(round(0.20 * n))
        # ensure non-empty calibration/test when possible
        if n >= 3:
            n_train = max(1, min(n - 2, n_train))
            n_cal = max(1, min(n - n_train - 1, n_cal))
        n_test = max(0, n - n_train - n_cal)
        train_ids = ids[:n_train]
        cal_ids = ids[n_train : n_train + n_cal]
        test_ids = ids[n_train + n_cal : n_train + n_cal + n_test]
        for sid in train_ids:
            split_map[sid] = "train"
        for sid in cal_ids:
            split_map[sid] = "calibration"
        for sid in test_ids:
            split_map[sid] = "test"

    for row in rows:
        row["split"] = split_map.get(str(row["scenario_id"]), "test")
    return rows


__all__ = [
    "assign_scenario_splits",
    "attach_recoverability_targets",
    "attach_scenario_stability",
    "compute_truth_aware_metrics",
    "spearman_corr",
]



"""Core inference and calibrated identifiability outputs."""

from __future__ import annotations

import json
from collections import Counter
from pathlib import Path
from typing import Any, Dict, List, Optional, Sequence, Set, Tuple

import numpy as np
import torch

from babappai.calibration import (
    D_OBS_DEFINITION,
    apply_ceii_calibration,
    get_neutral_reference,
    load_calibration_asset,
)
from babappai.dispersion import PRIMARY_DISPERSION_METHOD, compute_dispersion
from babappai.encoding import encode_alignment
from babappai.identifiability import eii01_from_eiiz
from babappai.metadata import MODEL_DOI, MODEL_FILE_NAME, MODEL_SHA256, MODEL_TAG
from babappai.model_manager import ensure_model, model_status
from babappai.stats import empirical_monte_carlo_pvalue
from babappai.tree import enumerate_branches, load_tree
from babappai.utils import resolve_device


def _is_terminal_branch(name: str) -> bool:
    return not name.startswith("internal_")


def _set_seed(seed: Optional[int]) -> None:
    if seed is None:
        return
    np.random.seed(seed)
    torch.manual_seed(seed)
    if torch.cuda.is_available():
        torch.cuda.manual_seed_all(seed)


def _load_foreground_list(path: Path) -> List[str]:
    entries: List[str] = []
    for line in path.read_text().splitlines():
        value = line.strip()
        if not value or value.startswith("#"):
            continue
        entries.append(value)
    return entries


def _select_foregrounds(
    branches: Sequence[str],
    foreground_mode: str,
    foreground_list_path: Optional[str],
) -> Set[str]:
    branch_set = set(branches)

    if foreground_mode == "all-leaves":
        return {b for b in branches if _is_terminal_branch(b)}

    if foreground_mode == "all-branches":
        return set(branches)

    if foreground_mode != "user-list":
        raise ValueError(f"Unsupported foreground mode: {foreground_mode}")

    if not foreground_list_path:
        raise ValueError(
            "foreground_mode=user-list requires --foreground-list to be set."
        )

    selected = _load_foreground_list(Path(foreground_list_path))
    if not selected:
        raise ValueError("Foreground list is empty.")

    unknown = sorted({item for item in selected if item not in branch_set})
    if unknown:
        raise ValueError(
            "Foreground list contains unknown branch IDs: " + ", ".join(unknown)
        )

    return set(selected)


def _extract_site_statistics(site_logits: torch.Tensor) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    squeezed = site_logits.squeeze(0)

    if squeezed.ndim == 1:
        site_logit_mean = squeezed
        site_logit_var = torch.zeros_like(squeezed)
        site_scores = torch.sigmoid(squeezed)
    elif squeezed.ndim == 2:
        site_logit_mean = squeezed.mean(dim=0)
        site_logit_var = squeezed.var(dim=0, unbiased=False)
        site_scores = torch.sigmoid(squeezed).mean(dim=0)
    else:
        raise ValueError(
            f"Unexpected site_logits dimensionality: {tuple(squeezed.shape)}"
        )

    return (
        site_scores.detach().cpu().numpy(),
        site_logit_mean.detach().cpu().numpy(),
        site_logit_var.detach().cpu().numpy(),
    )


def _extract_branch_logits(branch_logits: torch.Tensor, n_branches: int) -> np.ndarray:
    squeezed = branch_logits.squeeze(0)

    if squeezed.ndim == 1:
        branch_values = squeezed
    elif squeezed.ndim == 2:
        if squeezed.shape[0] == n_branches:
            branch_values = squeezed.mean(dim=1)
        elif squeezed.shape[1] == n_branches:
            branch_values = squeezed.mean(dim=0)
        else:
            branch_values = squeezed.reshape(-1)
    else:
        branch_values = squeezed.reshape(-1)

    if branch_values.numel() != n_branches:
        raise ValueError(
            "Model branch output size mismatch: "
            f"expected {n_branches}, got {branch_values.numel()}"
        )

    return branch_values.detach().cpu().numpy()


def _force_ceii_abstention(
    payload: Dict[str, Any],
    *,
    reason: str,
    status_override: Optional[str] = None,
) -> Dict[str, Any]:
    status = str(status_override or payload.get("applicability_status") or "calibration_unavailable")
    out = dict(payload)
    out.update(
        {
            "ceii_gene": None,
            "ceii_site": None,
            "ceii_gene_class": "calibration_unavailable",
            "ceii_site_class": "calibration_unavailable",
            "ceii_gene_identifiable_bool": False,
            "ceii_site_identifiable_bool": False,
            "ceii_ci": {
                "gene": {"lower": None, "upper": None},
                "site": {"lower": None, "upper": None},
            },
            "applicability_status": status,
            "domain_shift_or_applicability": status,
            "calibration_unavailable_reason": str(reason),
        }
    )
    return out


def run_inference(
    alignment_path: str,
    tree_path: Optional[str] = None,
    tree_obj=None,
    out_path: Optional[str] = None,
    device: str = "auto",
    model_tag: str = MODEL_TAG,
    tree_calibration: bool = False,
    N_calibration: int = 200,
    offline: bool = False,
    seed: Optional[int] = None,
    foreground_mode: str = "all-leaves",
    foreground_list_path: Optional[str] = None,
    batch_size: int = 1,
    sigma_floor: float = 0.0,
    alpha: float = 0.05,
    pvalue_mode: str = "empirical_monte_carlo",
    dispersion_method: str = PRIMARY_DISPERSION_METHOD,
    neutral_reps: int = 200,
    min_neutral_group_size: int = 20,
    retain_eii_bands: bool = True,
    report_threshold_bands: bool = True,
    ceii_enabled: bool = True,
    ceii_asset_path: Optional[str] = None,
    _model_override=None,
    _resolved_device_override=None,
) -> Dict[str, object]:
    """Run branch-conditioned inference and return structured outputs."""

    warnings: List[str] = [
        "Diagnostic identifiability score, not evidence of adaptive substitution.",
    ]
    if not (0.0 < alpha <= 1.0):
        raise ValueError("alpha must be in (0, 1].")
    if sigma_floor < 0:
        raise ValueError("sigma_floor must be >= 0.")
    if neutral_reps < 0:
        raise ValueError("neutral_reps must be >= 0.")
    if min_neutral_group_size < 0:
        raise ValueError("min_neutral_group_size must be >= 0.")
    allowed_modes = {"empirical_monte_carlo", "frozen_reference", "disabled"}
    if pvalue_mode not in allowed_modes:
        raise ValueError(f"Unsupported pvalue_mode: {pvalue_mode}. Allowed: {sorted(allowed_modes)}")

    _set_seed(seed)

    resolved_device = _resolved_device_override if _resolved_device_override is not None else resolve_device(device)
    if _model_override is None:
        model_path = ensure_model(model_tag, offline=offline)
        model = torch.jit.load(str(model_path), map_location=resolved_device)
    else:
        model = _model_override
        model_path = Path(model_status().get("cached_path") or "")
    model.eval()

    X, ntaxa, L = encode_alignment(alignment_path)

    tree = tree_obj if tree_obj is not None else load_tree(tree_path)
    branches = enumerate_branches(tree)
    K = len(branches)

    selected_foregrounds = _select_foregrounds(
        branches=branches,
        foreground_mode=foreground_mode,
        foreground_list_path=foreground_list_path,
    )

    taxon_names = [leaf.name for leaf in tree.get_leaves()]
    taxon_to_idx = {name: idx for idx, name in enumerate(taxon_names)}

    branch_to_taxa: List[List[int]] = []
    for node in tree.traverse("preorder"):
        if node.is_root():
            continue
        descendants = [
            taxon_to_idx[leaf.name]
            for leaf in node.get_leaves()
            if leaf.name in taxon_to_idx
        ]
        branch_to_taxa.append(descendants)

    parent = torch.zeros((1, K, L), dtype=torch.long, device=resolved_device)
    child = torch.zeros((1, K, L), dtype=torch.long, device=resolved_device)

    for i in range(L):
        col = X[:, i].tolist()
        consensus = Counter(col).most_common(1)[0][0]

        for b, taxa in enumerate(branch_to_taxa):
            if not taxa:
                continue
            mismatches = sum(X[t, i] != consensus for t in taxa)
            parent[0, b, i] = len(taxa) - mismatches
            child[0, b, i] = mismatches

    branch_length = torch.ones((1, K), dtype=torch.float32, device=resolved_device)

    with torch.no_grad():
        site_logits, branch_logits = model(parent, child, branch_length)

    site_scores, site_logit_mean, site_logit_var = _extract_site_statistics(site_logits)
    branch_logit_mean = _extract_branch_logits(branch_logits, K)
    branch_background = 1.0 / (1.0 + np.exp(-np.clip(branch_logit_mean, -60.0, 60.0)))

    sigma2_obs = compute_dispersion(site_logit_mean, method=str(dispersion_method))

    if batch_size != 1:
        warnings.append(
            "batch_size is currently accepted for CLI compatibility but not used by "
            "the frozen model runner."
        )

    calibration_group = f"L_{L}_K_{K}"
    calibration_source = "disabled_for_neutral_simulation"
    neutral_expected_variance: Optional[float]
    neutral_sd_raw: Optional[float]
    neutral_sd_floored: Optional[float]
    neutral_group_size: Optional[int] = None
    neutral_replicates: List[float] = []
    fallback_flag = False
    fallback_reason: Optional[str] = None
    p_emp_raw = float("nan")

    if pvalue_mode == "disabled":
        neutral_expected_variance = sigma2_obs
        neutral_sd_raw = 0.0
        neutral_group_size = 0
        p_emp_raw = 1.0
    elif pvalue_mode == "empirical_monte_carlo":
        from babappai.tree_calibration import monte_carlo_neutral

        n_empirical = int(neutral_reps if neutral_reps > 0 else N_calibration)
        if n_empirical <= 0:
            raise ValueError("At least one neutral replicate is required for empirical_monte_carlo mode.")
        calibration_source = "tree_conditional_empirical_monte_carlo"
        mu0, sd0, sims = monte_carlo_neutral(
            tree=tree,
            L=L,
            inference_function=run_inference,
            model_tag=model_tag,
            N=n_empirical,
            offline=offline,
            foreground_mode=foreground_mode,
            foreground_list_path=foreground_list_path,
            batch_size=batch_size,
            inference_extra_kwargs={
                "_model_override": model,
                "_resolved_device_override": resolved_device,
                "retain_eii_bands": retain_eii_bands,
                "report_threshold_bands": report_threshold_bands,
                "dispersion_method": str(dispersion_method),
            },
        )
        sims_arr = np.asarray(sims, dtype=float)
        neutral_replicates = sims_arr.tolist()
        neutral_expected_variance = float(mu0)
        neutral_sd_raw = float(sd0)
        neutral_group_size = int(sims_arr.size)
        p_emp_raw = empirical_monte_carlo_pvalue(sigma2_obs, neutral_replicates)
    else:
        calibration_source = "frozen_reference_table"
        reference = get_neutral_reference(L=L, K=K, model_tag=model_tag)
        if reference is None:
            neutral_expected_variance = sigma2_obs
            neutral_sd_raw = None
            neutral_group_size = 0
            fallback_flag = True
            fallback_reason = "missing_neutral_reference"
            calibration_source = "fallback_uncalibrated"
            warnings.append(
                "No frozen neutral reference available for this alignment/tree size; "
                "falling back to floor-bounded sigma0 with p_emp unavailable."
            )
        else:
            neutral_expected_variance = float(reference["sigma2_mean"])
            neutral_sd_raw = float(reference["sigma2_sd"])
            if "n_replicates" in reference:
                neutral_group_size = int(reference["n_replicates"])
            elif "n" in reference:
                neutral_group_size = int(reference["n"])
            else:
                neutral_group_size = None
        warnings.append(
            "pvalue_mode=frozen_reference does not provide empirical Monte Carlo p-values; "
            "use pvalue_mode=empirical_monte_carlo for significance decisions."
        )

    if pvalue_mode == "empirical_monte_carlo" and int(neutral_group_size or 0) < int(min_neutral_group_size):
        fallback_flag = True
        if fallback_reason is None:
            fallback_reason = "neutral_group_below_minimum"
        warnings.append(
            "Neutral calibration group size below configured minimum "
            f"({neutral_group_size} < {min_neutral_group_size})."
        )

    sigma0_raw_finite = bool(neutral_sd_raw is not None and np.isfinite(neutral_sd_raw))
    if not sigma0_raw_finite or float(neutral_sd_raw) <= 0:
        if not fallback_flag:
            fallback_flag = True
            fallback_reason = "non_positive_or_invalid_neutral_sd"
        neutral_sd_base = 0.0
    else:
        neutral_sd_base = float(neutral_sd_raw)

    neutral_sd_floored = max(neutral_sd_base, float(sigma_floor))
    if neutral_sd_floored <= 0:
        fallback_flag = True
        if fallback_reason is None:
            fallback_reason = "non_positive_sigma_after_floor"
        neutral_sd_floored = 1e-8

    sigma_floor_applied = bool(sigma_floor > 0 and neutral_sd_floored <= float(sigma_floor) + 1e-12)
    sigma0_floored = bool(sigma_floor_applied)
    fallback_applied = bool(fallback_flag)

    sigma0_valid = bool(
        pvalue_mode != "disabled"
        and sigma0_raw_finite
        and float(neutral_sd_raw) > 0
        and not sigma0_floored
        and not fallback_applied
    )

    dispersion_ratio = None
    if neutral_expected_variance and neutral_expected_variance > 0:
        dispersion_ratio = float(sigma2_obs / neutral_expected_variance)

    eii_z_raw = float((sigma2_obs - neutral_expected_variance) / neutral_sd_floored)
    eii_01_raw = float(eii01_from_eiiz(eii_z_raw))
    if np.isfinite(p_emp_raw):
        q_emp = float(p_emp_raw)
    else:
        q_emp = float("nan")
    significant_bool = bool(np.isfinite(q_emp) and q_emp <= float(alpha))
    significance_label = "significant" if significant_bool else "not_significant"

    if fallback_flag and fallback_reason:
        warnings.append(
            "Neutral calibration fallback applied: "
            f"{fallback_reason}; using sigma0={neutral_sd_floored:.6g}."
        )
    ceii_payload: Dict[str, Any] = {
        "ceii_gene": None,
        "ceii_site": None,
        "ceii_gene_class": "calibration_unavailable",
        "ceii_site_class": "calibration_unavailable",
        "ceii_gene_identifiable_bool": False,
        "ceii_site_identifiable_bool": False,
        "ceii_ci": {
            "gene": {"lower": None, "upper": None},
            "site": {"lower": None, "upper": None},
        },
        "applicability_score": None,
        "applicability_status": "calibration_unavailable",
        "within_applicability_envelope": False,
        "calibration_unavailable_reason": "calibration_not_evaluated",
        "nearest_supported_regime": "unknown",
        "distance_to_supported_domain": None,
        "domain_shift_or_applicability": "calibration_unavailable",
        "calibration_version": "unavailable",
    }
    if ceii_enabled:
        try:
            calibration_asset = load_calibration_asset(ceii_asset_path)
            ceii_payload = apply_ceii_calibration(
                eii_z_raw=eii_z_raw,
                n_taxa=ntaxa,
                gene_length_nt=L,
                n_branches=K,
                asset=calibration_asset,
            )
        except Exception as exc:
            warnings.append(
                "cEII calibration unavailable; returning raw EII diagnostics only. "
                f"Reason: {exc}"
            )
            ceii_payload = _force_ceii_abstention(
                ceii_payload,
                reason=f"calibration_runtime_error:{exc}",
                status_override="calibration_unavailable",
            )
    else:
        warnings.append("cEII calibration disabled by caller; returning raw EII diagnostics only.")
        ceii_payload = _force_ceii_abstention(
            ceii_payload,
            reason="ceii_disabled_by_caller",
            status_override="calibration_unavailable",
        )

    sigma_abstain_reasons: List[str] = []
    if pvalue_mode == "disabled":
        sigma_abstain_reasons.append("null_calibration_disabled")
    if not sigma0_raw_finite:
        sigma_abstain_reasons.append("sigma0_non_finite")
    elif float(neutral_sd_raw) <= 0:
        sigma_abstain_reasons.append("sigma0_non_positive")
    if sigma0_floored:
        sigma_abstain_reasons.append("sigma0_floored")
    if fallback_applied:
        sigma_abstain_reasons.append(
            "null_calibration_fallback_applied"
            + (f":{fallback_reason}" if fallback_reason else "")
        )

    if sigma_abstain_reasons:
        existing_reason = str(ceii_payload.get("calibration_unavailable_reason", "") or "")
        sigma_reason = ",".join(sigma_abstain_reasons)
        combined_reason = (
            f"{existing_reason}; {sigma_reason}"
            if existing_reason and sigma_reason not in existing_reason
            else sigma_reason or existing_reason
        )
        current_status = str(ceii_payload.get("applicability_status", "") or "")
        status_override = (
            "calibration_unavailable"
            if current_status in {"in_domain", "near_boundary", "unknown", ""}
            else None
        )
        ceii_payload = _force_ceii_abstention(
            ceii_payload,
            reason=combined_reason,
            status_override=status_override,
        )

    if retain_eii_bands:
        band_extent = str(ceii_payload.get("ceii_gene_class", "calibration_unavailable"))
        band_bool = bool(ceii_payload.get("ceii_gene_identifiable_bool", False))
    else:
        band_extent = "descriptive_band_suppressed"
        band_bool = False

    calib_reason = str(ceii_payload.get("calibration_unavailable_reason", "") or "")
    if calib_reason:
        warnings.append(
            "cEII probability abstained by applicability policy: "
            f"{calib_reason}."
        )

    warnings.append(
        "Raw EII is a dispersion magnitude statistic. Decision-ready identifiability "
        "is represented by cEII outputs calibrated on held-out simulation truth."
    )
    warnings.append(
        "Inferential significance remains separate: use p_emp/q_emp/significant_bool "
        "for matched-neutral exceedance support."
    )
    gene_summary = {
        "D_obs": sigma2_obs,
        "D_obs_definition": D_OBS_DEFINITION,
        "D_obs_method": str(dispersion_method),
        "mu0": neutral_expected_variance,
        "sigma0_raw": neutral_sd_raw,
        "sigma0_final": neutral_sd_floored,
        "sigma_floor_used": float(sigma_floor),
        "sigma0_valid": sigma0_valid,
        "sigma0_floored": sigma0_floored,
        "fallback_applied": fallback_applied,
        "fallback_flag": fallback_flag,
        "fallback_reason": fallback_reason,
        "neutral_group_size": neutral_group_size,
        "calibration_group": calibration_group,
        "calibration_source": calibration_source,
        "p_emp": p_emp_raw,
        "p_emp_raw": p_emp_raw,
        "q_emp": q_emp,
        "alpha_used": float(alpha),
        "significant_bool": significant_bool,
        "significance_label": significance_label,
        "pvalue_mode": pvalue_mode,
        "neutral_replicates": neutral_replicates,
        "eii_band_descriptive_only": False,
        "retain_eii_bands": bool(retain_eii_bands),
        "report_threshold_bands": bool(report_threshold_bands),
        "observed_variance": sigma2_obs,
        "neutral_expected_variance": neutral_expected_variance,
        "neutral_sd": neutral_sd_floored,
        "neutral_sd_raw": neutral_sd_raw,
        "neutral_sd_floored": neutral_sd_floored,
        "sigma_floor": float(sigma_floor),
        "sigma_floor_applied": sigma_floor_applied,
        "calibration_fallback_flag": fallback_flag,
        "calibration_fallback_reason": fallback_reason,
        "dispersion_ratio": dispersion_ratio,
        "eii_z_raw": eii_z_raw,
        "eii_01_raw": eii_01_raw,
        "EII_z": eii_z_raw,
        "EII_01": eii_01_raw,
        "ceii_gene": ceii_payload.get("ceii_gene"),
        "ceii_site": ceii_payload.get("ceii_site"),
        "ceii_gene_class": str(ceii_payload["ceii_gene_class"]),
        "ceii_site_class": str(ceii_payload["ceii_site_class"]),
        "ceii_gene_identifiable_bool": bool(ceii_payload["ceii_gene_identifiable_bool"]),
        "ceii_site_identifiable_bool": bool(ceii_payload["ceii_site_identifiable_bool"]),
        "ceii_ci": ceii_payload["ceii_ci"],
        "applicability_score": ceii_payload.get("applicability_score"),
        "applicability_status": str(ceii_payload.get("applicability_status", "unknown")),
        "within_applicability_envelope": bool(ceii_payload.get("within_applicability_envelope", False)),
        "calibration_unavailable_reason": ceii_payload.get("calibration_unavailable_reason"),
        "nearest_supported_regime": str(ceii_payload.get("nearest_supported_regime", "unknown")),
        "distance_to_supported_domain": ceii_payload.get("distance_to_supported_domain"),
        "domain_shift_or_applicability": str(ceii_payload["domain_shift_or_applicability"]),
        "calibration_version": str(ceii_payload["calibration_version"]),
        "identifiable_bool": band_bool,
        "identifiability_extent": band_extent,
        "identifiable_bool_deprecated": band_bool,
        "identifiability_extent_deprecated": band_extent,
        "model_version": str(model_tag),
        "model_checkpoint_provenance": {
            "model_file_name": MODEL_FILE_NAME,
            "model_doi": MODEL_DOI,
            "model_sha256": MODEL_SHA256,
            "cached_path": str(model_path),
        },
    }

    branch_results = []
    for idx, branch in enumerate(branches):
        branch_results.append(
            {
                "branch": branch,
                "background_score": float(branch_background[idx]),
                "logit_mean": float(branch_logit_mean[idx]),
                "is_terminal": _is_terminal_branch(branch),
                "selected_foreground": branch in selected_foregrounds,
            }
        )

    site_results = []
    for idx in range(L):
        site_results.append(
            {
                "site": idx + 1,
                "site_score": float(site_scores[idx]),
                "site_logit_mean": float(site_logit_mean[idx]),
                "site_logit_var": float(site_logit_var[idx]),
            }
        )

    result = {
        "model": "BABAPPAiLegacyFrozenModel",
        "model_tag": model_tag,
        "model_version": str(model_tag),
        "model_checkpoint_provenance": {
            "model_file_name": MODEL_FILE_NAME,
            "model_doi": MODEL_DOI,
            "model_sha256": MODEL_SHA256,
            "cached_path": str(model_path),
        },
        "num_branches": K,
        "num_sites": L,
        "n_sequences": ntaxa,
        "device": str(resolved_device),
        "model_status": {**model_status(), "cached_path": str(model_path), "verified": True},
        "site_scores": site_scores.tolist(),
        "site_logit_mean": site_logit_mean.tolist(),
        "site_logit_var": site_logit_var.tolist(),
        "branch_background": {
            branch: float(score)
            for branch, score in zip(branches, branch_background.tolist())
        },
        "branch_logit_mean": {
            branch: float(score)
            for branch, score in zip(branches, branch_logit_mean.tolist())
        },
        "gene_level_identifiability": gene_summary,
        "branch_results": branch_results,
        "site_results": site_results,
        "foreground": {
            "mode": foreground_mode,
            "selected_branches": sorted(selected_foregrounds),
        },
        "neutral_calibration_replicates": neutral_replicates,
        "warnings": warnings,
    }

    if out_path:
        Path(out_path).write_text(json.dumps(result, indent=2) + "\n")

    return result



from __future__ import annotations

from babappai.run_pipeline import terminal_summary


def test_terminal_summary_reports_calibration_withheld_warning() -> None:
    payload = {
        "gene_summary": {
            "p_emp": 0.2,
            "q_emp": 0.2,
            "alpha_used": 0.05,
            "significant_bool": False,
            "eii_z_raw": 0.1,
            "eii_01_raw": 0.52,
            "ceii_gene": None,
            "ceii_site": None,
            "ceii_gene_class": "calibration_unavailable",
            "calibration_version": "ceii_v2",
            "applicability_status": "calibration_unavailable",
            "applicability_score": 1.0,
            "calibration_unavailable_reason": "sigma0_floored",
            "sigma0_valid": False,
            "sigma0_floored": True,
            "fallback_applied": False,
        },
        "branch_results": [
            {"branch": "a", "background_score": 0.2},
            {"branch": "b", "background_score": 0.1},
        ],
        "site_results": [
            {"site": 1, "site_score": 0.8},
            {"site": 2, "site_score": 0.7},
        ],
    }
    lines = terminal_summary(payload)
    assert any("calibrated ceii outputs withheld" in line.lower() for line in lines)
    assert any("sigma0_valid=False" in line for line in lines)



from __future__ import annotations

from pathlib import Path

import numpy as np
import torch

import babappai.inference as inference
import babappai.tree_calibration as tree_calibration
from babappai.stats import bh_adjust, empirical_monte_carlo_pvalue


class _FakeLeaf:
    def __init__(self, name: str):
        self.name = name

    def is_root(self) -> bool:
        return False

    def get_leaves(self):
        return [self]


class _FakeRoot:
    def __init__(self, leaves):
        self._leaves = leaves

    def is_root(self) -> bool:
        return True

    def get_leaves(self):
        return list(self._leaves)

    def traverse(self, _order: str):
        return [self, *self._leaves]


class _FakeModel:
    def eval(self):
        return self

    def __call__(self, parent, child, branch_length):
        _ = parent, child, branch_length
        site_logits = torch.tensor([[0.10, 0.40, -0.20, 0.00]], dtype=torch.float32)
        branch_logits = torch.tensor([[0.20, -0.10]], dtype=torch.float32)
        return site_logits, branch_logits


def _patch_inference_runtime(monkeypatch):
    leaves = [_FakeLeaf("taxonA"), _FakeLeaf("taxonB")]
    root = _FakeRoot(leaves)
    x = np.asarray([[1, 2, 3, 1], [1, 2, 0, 1]], dtype=int)

    monkeypatch.setattr(inference, "encode_alignment", lambda _: (x, 2, 4))
    monkeypatch.setattr(inference, "load_tree", lambda _: root)
    monkeypatch.setattr(inference, "enumerate_branches", lambda _: ["taxonA", "taxonB"])
    monkeypatch.setattr(inference, "ensure_model", lambda *args, **kwargs: Path("/tmp/fake_model.pt"))
    monkeypatch.setattr(inference, "model_status", lambda: {"cached_path": "/tmp/fake_model.pt", "verified": True})
    monkeypatch.setattr(inference.torch.jit, "load", lambda *args, **kwargs: _FakeModel())


def test_empirical_pvalue_formula():
    observed = 0.6
    neutral = [0.1, 0.3, 0.7, 0.8]
    p_emp = empirical_monte_carlo_pvalue(observed, neutral)
    expected = (1 + 2) / (4 + 1)  # 0.7 and 0.8 exceed/meet observed
    assert abs(p_emp - expected) < 1e-12


def test_bh_adjust_sanity():
    pvals = np.asarray([0.01, 0.04, 0.03, 0.20], dtype=float)
    qvals = bh_adjust(pvals)
    expected = np.asarray([0.04, 0.0533333333333, 0.0533333333333, 0.20], dtype=float)
    assert np.allclose(qvals, expected, atol=1e-10)


def test_inference_empirical_significance_and_sigma_floor(monkeypatch):
    _patch_inference_runtime(monkeypatch)

    neutral = np.asarray([0.01, 0.03, 0.08, 0.10], dtype=float)
    mu0 = float(np.mean(neutral))
    sd0 = float(np.std(neutral, ddof=1))
    monkeypatch.setattr(
        tree_calibration,
        "monte_carlo_neutral",
        lambda **kwargs: (mu0, sd0, neutral),
    )
    monkeypatch.setattr(
        inference,
        "load_calibration_asset",
        lambda *_args, **_kwargs: {
            "calibration_version": "ceii_test",
            "gene_calibrator": {"x": [-3.0, 0.0, 3.0], "y": [0.1, 0.5, 0.9]},
            "site_calibrator": {"x": [-3.0, 0.0, 3.0], "y": [0.05, 0.4, 0.8]},
            "thresholds": {
                "gene": {"threshold": 0.6},
                "site": {"threshold": 0.65},
            },
            "classes": {
                "gene": [
                    {"label": "not_identifiable", "min": 0.0, "max": 0.4},
                    {"label": "weak_or_ambiguous", "min": 0.4, "max": 0.6},
                    {"label": "identifiable", "min": 0.6, "max": 0.8},
                    {"label": "strongly_identifiable", "min": 0.8, "max": 1.0},
                ],
                "site": [
                    {"label": "not_identifiable", "min": 0.0, "max": 0.4},
                    {"label": "weak_or_ambiguous", "min": 0.4, "max": 0.65},
                    {"label": "identifiable", "min": 0.65, "max": 0.8},
                    {"label": "strongly_identifiable", "min": 0.8, "max": 1.0},
                ],
            },
            "applicability": {
                "features": {
                    "n_taxa": {"min": 2, "max": 128},
                    "gene_length_nt": {"min": 3, "max": 10000},
                },
                "near_boundary_fraction": 0.01,
                "min_applicability_score_for_calibration": 0.0,
                "allow_near_boundary_calibration": True,
            },
        },
    )

    result = inference.run_inference(
        alignment_path="dummy.fasta",
        tree_path="dummy.nwk",
        offline=True,
        pvalue_mode="empirical_monte_carlo",
        neutral_reps=4,
        min_neutral_group_size=1,
        sigma_floor=0.05,
        alpha=0.05,
        seed=7,
    )
    gene = result["gene_level_identifiability"]
    required_schema = {
        "applicability_score",
        "applicability_status",
        "within_applicability_envelope",
        "calibration_unavailable_reason",
        "nearest_supported_regime",
        "distance_to_supported_domain",
        "sigma0_valid",
        "sigma0_floored",
        "fallback_applied",
        "ceii_gene",
        "ceii_site",
        "ceii_gene_class",
        "ceii_site_class",
        "eii_z_raw",
        "eii_01_raw",
        "q_emp",
        "significant_bool",
    }
    assert required_schema.issubset(gene.keys())

    expected_p = empirical_monte_carlo_pvalue(gene["D_obs"], neutral.tolist())
    assert abs(float(gene["p_emp"]) - float(expected_p)) < 1e-12
    assert gene["neutral_replicates"] == neutral.tolist()
    assert float(gene["sigma0_final"]) >= 0.05
    assert abs(float(gene["sigma0_final"]) - float(gene["neutral_sd"])) < 1e-12
    assert bool(gene["sigma0_floored"]) is True
    assert bool(gene["sigma0_valid"]) is False
    assert bool(gene["fallback_applied"]) is False
    assert bool(gene["significant_bool"]) == (float(gene["q_emp"]) <= 0.05)
    assert float(gene["EII_z"]) == float(gene["eii_z_raw"])
    assert 0.0 <= float(gene["eii_01_raw"]) <= 1.0
    assert gene["ceii_gene"] is None
    assert gene["ceii_site"] is None
    assert gene["ceii_gene_class"] == "calibration_unavailable"
    assert str(gene["applicability_status"]) == "calibration_unavailable"
    assert "sigma0_floored" in str(gene.get("calibration_unavailable_reason", ""))
    assert bool(gene["within_applicability_envelope"]) is True
    assert gene["identifiability_extent"] == gene["ceii_gene_class"]


def test_inference_abstains_when_sigma_nonfinite(monkeypatch):
    _patch_inference_runtime(monkeypatch)

    neutral = np.asarray([0.01, 0.02, 0.03, 0.04], dtype=float)
    mu0 = float(np.mean(neutral))
    monkeypatch.setattr(
        tree_calibration,
        "monte_carlo_neutral",
        lambda **kwargs: (mu0, float("nan"), neutral),
    )
    monkeypatch.setattr(
        inference,
        "load_calibration_asset",
        lambda *_args, **_kwargs: {
            "calibration_version": "ceii_test",
            "gene_calibrator": {"x": [-3.0, 0.0, 3.0], "y": [0.1, 0.5, 0.9]},
            "site_calibrator": {"x": [-3.0, 0.0, 3.0], "y": [0.05, 0.4, 0.8]},
            "thresholds": {
                "gene": {"threshold": 0.6},
                "site": {"threshold": 0.65},
            },
            "classes": {
                "gene": [
                    {"label": "not_identifiable", "min": 0.0, "max": 0.4},
                    {"label": "weak_or_ambiguous", "min": 0.4, "max": 0.6},
                    {"label": "identifiable", "min": 0.6, "max": 0.8},
                    {"label": "strongly_identifiable", "min": 0.8, "max": 1.0},
                ],
                "site": [
                    {"label": "not_identifiable", "min": 0.0, "max": 0.4},
                    {"label": "weak_or_ambiguous", "min": 0.4, "max": 0.65},
                    {"label": "identifiable", "min": 0.65, "max": 0.8},
                    {"label": "strongly_identifiable", "min": 0.8, "max": 1.0},
                ],
            },
            "applicability": {
                "features": {
                    "n_taxa": {"min": 2, "max": 128},
                    "gene_length_nt": {"min": 3, "max": 10000},
                },
                "near_boundary_fraction": 0.01,
                "min_applicability_score_for_calibration": 0.0,
                "allow_near_boundary_calibration": True,
            },
        },
    )
    result = inference.run_inference(
        alignment_path="dummy.fasta",
        tree_path="dummy.nwk",
        offline=True,
        pvalue_mode="empirical_monte_carlo",
        neutral_reps=4,
        sigma_floor=0.001,
        alpha=0.05,
        seed=11,
    )
    gene = result["gene_level_identifiability"]
    assert gene["ceii_gene"] is None
    assert gene["ceii_site"] is None
    assert gene["ceii_gene_class"] == "calibration_unavailable"
    assert str(gene["applicability_status"]) == "calibration_unavailable"
    assert "sigma0_non_finite" in str(gene.get("calibration_unavailable_reason", ""))
    assert bool(gene["sigma0_valid"]) is False
